from ..lc_archive import LcArchive
from ...from_json import from_json

import os
import json
from shutil import rmtree
import unittest
from datetime import datetime

work_dir = os.path.join(os.path.dirname(__file__), 'scratch')
test_file = os.path.join(os.path.dirname(__file__), 'test_json.json')

test_json = {
 'dataReference': 'test.basic',
 'dataSource': None,
 'flows': [
     {
         'CasNumber': '',
         'Comment': '',
         'Compartment': ['Forestry and Logging', 'Logging'],
         'Name': 'Reforesting, average state or private moist cold softwood forest, INW',
         'characterizations': [{
             'entityType': 'characterization',
             'isReference': True,
             'quantity': 'cec3a58d-44c3-31f6-9c75-90e6352f0934',
             'value': 1.0}],
         'entityId': '51d92a19-e550-3000-b49b-e633137eac06',
         'entityType': 'flow',
         'externalId': 37572,
         'origin': 'local.uslci.clean'
     },
     {
         'CasNumber': '',
         'Comment': '',
         'Compartment': ['Forestry and Logging', 'Forest Nurseries, Gathering Forest Prod.'],
         'Name': 'Greenhouse seedling, softwood, INW',
         'characterizations': [{
             'entityType': 'characterization',
             'isReference': True,
             'quantity': '8703965a-7a6b-3e3e-a1cf-d9adf7bf1d9f',
             'value': 1.0}],
         'entityId': '9cc0ccce-8e33-35ca-a3c0-c7bb6c397e95',
         'entityType': 'flow',
         'externalId': 24235,
         'origin': 'local.uslci.clean'
     },
     {
         'CasNumber': '',
         'Comment': '',
         'Compartment': ['Utilities', 'Steam and Air-Conditioning Supply'],
         'Name': 'Gasoline, combusted in equipment',
         'characterizations': [{
             'entityType': 'characterization',
             'isReference': True,
             'quantity': '21d34f33-b0af-3d82-9bef-3cf03e0db9dc',
             'value': 1.0}],
         'entityId': 'c65e7555-7652-377b-b69a-126ba3d9b8b7',
         'entityType': 'flow',
         'externalId': 5233,
         'origin': 'local.uslci.clean'
     }],
    'initArgs':
        {
            'ns_uuid': 'a9d158e0-d48c-4427-8c55-42719e9e11cc'
        },
    'processes': [
     {
         'Classifications': ['Forestry and Logging', 'Logging'],
         'Comment': 'Important note: although most of the data in the US LCI database has undergone some sort of \
                    review, the database as a whole has not yet undergone a formal validation process. Please email \
                    comments to lci@nrel.gov.',
         'Name': 'Reforesting, average state or private moist cold softwood forest, INW',
         'SpatialScope': 'RNA',
         'TemporalScope': "{'begin': '1989-01-01-07:00', 'end': '1996-01-01-07:00'}",
         'entityId': '78c8b1e5-ca60-38b6-9a94-dde046560a38',
         'entityType': 'process',
         'exchanges': [
             {
                 'direction': 'Input',
                 'entityType': 'exchange',
                 'flow': '9cc0ccce-8e33-35ca-a3c0-c7bb6c397e95',
                 'value': 865.0
             },
             {
                 'direction': 'Input',
                 'entityType': 'exchange',
                 'flow': 'c65e7555-7652-377b-b69a-126ba3d9b8b7',
                 'value': 15.959999999999999
             },
             {
                 'direction': 'Output',
                 'entityType': 'exchange',
                 'flow': '51d92a19-e550-3000-b49b-e633137eac06',
                 'isReference': True,
                 'value': 1.0
             }],
         'externalId': 'Reforesting, average state or private moist cold softwood forest, INW',
         'origin': 'local.uslci.clean'
     }],
 'quantities': [
     {
         'Comment': 'EcoSpold01 with altered uuid',
         'Name': 'EcoSpold Quantity l',
         'entityId': '21d34f33-c0af-3d82-9bef-3cf03e0db9dc',
         'entityType': 'quantity',
         'externalId': 'l',
         'origin': 'local.uslci.clean',
         'referenceUnit': 'l'
     },
     {
         'Comment': 'EcoSpold01',
         'Name': 'EcoSpold Quantity Item(s)',
         'entityId': '8703965a-7a6b-3e3e-a1cf-d9adf7bf1d9f',
         'entityType': 'quantity',
         'externalId': 'Item(s)',
         'origin': 'local.uslci.clean',
         'referenceUnit': 'Item(s)'
     },
     {
         'Comment': 'EcoSpold01',
         'Name': 'EcoSpold Quantity ha',
         'entityId': 'cec3a58d-44c3-31f6-9c75-90e6352f0934',
         'entityType': 'quantity',
         'externalId': 'ha',
         'origin': 'local.uslci.clean',
         'referenceUnit': 'ha'
     }]
}


def setUpModule():
    if not os.path.exists(test_file):
        with open(test_file, 'w') as fp:
            print('writing to %s' % test_file)
            json.dump(test_json, fp, indent=2, sort_keys=True)


class LcArchiveTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._ar = LcArchive.from_file(test_file)

    def test_get_none(self):
        """
        get None returns None
        :return:
        """
        self.assertIsNone(self._ar[None])

    def test_get_by_uuid(self):
        """
        the standard getitem: put in an entity key, get back the entity.
        :return:
        """
        uuid = '21d34f33-c0af-3d82-9bef-3cf03e0db9dc'
        nsuuid = self._ar._ref_to_nsuuid('l')
        ent = self._ar[uuid]
        self.assertEqual(ent.entity_type, 'quantity')
        self.assertEqual(ent.external_ref, 'l')
        self.assertNotEqual(ent.uuid, nsuuid)
        self.assertEqual(self._ar[nsuuid].uuid, uuid)

    def test_get_by_nsuuid(self):
        uuid = 'cec3a58d-44c3-31f6-9c75-90e6352f0934'
        ent = self._ar['ha']
        self.assertEqual(ent.uuid, uuid)

    def test_get_entity(self):
        """
        An error-preventer: put in an entity itself, get back the entity IF the entity is part of the archive.
        :return:
        """
        ent = self._ar['21d34f33-b0af-3d82-9bef-3cf03e0db9dc']
        self.assertIs(self._ar[ent], ent)

    def test_ents_by_type(self):
        self.assertEqual(len([q for q in self._ar.entities_by_type('q')]), 3)

    def test_recursive_references(self):
        fl_uuid = '9cc0ccce-8e33-35ca-a3c0-c7bb6c397e95'
        q_uuid = '8703965a-7a6b-3e3e-a1cf-d9adf7bf1d9f'
        fl = self._ar[fl_uuid]
        self.assertEqual(self._ar._ref_to_nsuuid(fl.external_ref), fl_uuid)
        self.assertIs(self._ar[q_uuid], fl.reference_entity)

    def test_name(self):
        self.assertEqual(self._ar.ref, test_json['dataReference'])

    def test_json_init(self):
        """
        using load_from_dict(), you cannot read namespace uuid automatically from the json. But, you can specify source and
        ref manually.
        :return:
        """
        ar = LcArchive('/my/file')
        ar._entities['21d34f33-b0af-3d82-9bef-3cf03e0db9dc'] = None  # need to avoid KeyError with nsuuid
        ar.load_from_dict(from_json(test_file), jsonfile='/my/test/json')
        self.assertIn(test_json['dataReference'], ar.catalog_names)
        self.assertEqual(ar.ref, 'local.my.file')
        self.assertSequenceEqual([k for k in ar.get_sources(test_json['dataReference'])], [None])
        uuid = 'cec3a58d-44c3-31f6-9c75-90e6352f0934'
        nsuuid = ar._ref_to_nsuuid('ha')
        self.assertNotEqual(uuid, nsuuid)
        ent = ar['ha']
        ent1 = ar[nsuuid]
        self.assertIs(ent, ent1)
        self.assertEqual(ar[nsuuid].uuid, uuid)

    def test_assign_name(self):
        ar = LcArchive(test_json['dataSource'], ref='test.basic', ns_uuid=test_json['initArgs']['ns_uuid'])
        self.assertIsNone(ar.source)
        self.assertEqual(ar.ref, 'test.basic')
        ar.load_from_dict(from_json(test_file), jsonfile=test_file)
        self.assertEqual(ar.source, test_file)
        self.assertEqual(ar.ref, 'test.basic')


class DescendantTest(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        try:
            rmtree(work_dir)
        except FileNotFoundError:
            pass

    def test_create_descendant(self):
        new_date = datetime.now().strftime('%Y%m%d')
        sig = 'mybaby'

        ar = LcArchive.from_file(test_file)

        dref = ar.create_descendant(work_dir, signifier=sig, force=True)
        self.assertEqual(dref, ar.ref)
        self.assertTrue(dref.startswith('.'.join(['test.basic', sig, new_date])))
        self.assertSetEqual({k for k in ar.get_sources(dref)}, {os.path.join(work_dir, '%s.json.gz' % dref)})


if __name__ == '__main__':
    unittest.main()



from .ilcd import uuid_regex

from ...implementations import QuantityImplementation


class IlcdQuantityImplementation(QuantityImplementation):
    def get_canonical(self, quantity, **kwargs):
        """
        Retrieve a canonical quantity from a qdb
        :param quantity: external_id of quantity
        :return: quantity entity
        """
        u = uuid_regex.search(quantity).groups()[0]
        return self._archive.load_lcia_method(u, load_all_flows=False)

    def flowables(self, quantity=None, compartment=None, **kwargs):
        """
        Return a list of flowable strings. Use quantity and compartment parameters to narrow the result
        set to those characterized by a specific quantity, those exchanged with a specific compartment, or both
        :param quantity:
        :param compartment: filter by compartment not implemented
        :return: list of pairs: CAS number, name
        """
        fbs = set()
        if quantity is not None:
            for factor in self._archive.generate_factors(quantity):
                fb = factor.flow['CasNumber'], factor.flow['Name']
                if fb not in fbs:
                    fbs.add(fb)
                    yield fb
        else:
            for f in self._archive.entities_by_type('flow'):
                fb = f['CasNumber'], f['Name']
                if fb not in fbs:
                    fbs.add(fb)
                    yield fb

    def compartments(self, quantity=None, flowable=None, **kwargs):
        """
        Return a list of compartment strings. Use quantity and flowable parameters to narrow the result
        set to those characterized for a specific quantity, those with a specific flowable, or both
        :param quantity:
        :param flowable:
        :return: list of strings
        """
        pass

    def factors(self, quantity, flowable=None, context=None, **kwargs):
        """
        Return characterization factors for the given quantity, subject to optional flowable and compartment
        filter constraints. This is ill-defined because the reference unit is not explicitly reported in current
        serialization for characterizations (it is implicit in the flow)-- but it can be added to a web service layer.
        :param quantity:
        :param flowable:
        :param context: not implemented
        :return:
        """
        for factor in self._archive.generate_factors(quantity):
            if flowable is not None:
                if factor.flow['Name'] != flowable and factor.flow['CasNumber'] != flowable:
                    continue
            yield factor

    def quantity_relation(self, flowable, ref_quantity, query_quantity, context, locale='GLO', **kwargs):
        """
        Return a single number that converts the a unit of the reference quantity into the query quantity for the
        given flowable, compartment, and locale (default 'GLO').  If no locale is found, this would be a great place
        to run a spatial best-match algorithm.
        :param flowable:
        :param ref_quantity:
        :param query_quantity:
        :param context:
        :param locale:
        :return:
        """
        pass



from ..characterizations import Characterization

class QuelledCF(Characterization):

    @classmethod
    def from_cf(cls, cf, flowable=None):
        if flowable is None:
            flowable = cf.flowable
        quelled = cls(flowable, cf.ref_quantity, cf.quantity, cf.context, cf.origin)
        for l in cf.locations():
            quelled.add_value(value=0.0, location=l)
        return quelled

    @property
    def value(self):
        return 0.0

    @value.setter
    def value(self, val):
        raise TypeError('Cannot set value')

    def update_values(self, **kwargs):
        raise TypeError('Cannot set values')

    def __setitem__(self, item, value):
        self._locations[item] = 0.0

    def add_value(self, value=None, location=None, overwrite=False):
        super(QuelledCF, self).add_value(value=0.0, location=location)





import unittest

from .test_aa_local import cat
from ..local import make_config, check_enabled


_run_test = check_enabled('ipcc2007')


if _run_test:
    cfg = make_config('ipcc2007')
    ref = next(cfg.references)


class GwpIpcc2007Test(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.query = cat.query(ref)

    def test_resources_exist(self):
        self.assertIn(ref, cat.references)

    def test_num_entities(self):
        self.assertEqual(self.query.count('quantity'), 2)
        self.assertEqual(self.query.count('flow'), 0)

    def test_gwp(self):
        gwp = next(self.query.lcia_methods())
        self.assertEqual(gwp.name, 'Global Warming Air [kg CO2 eq] [LCIA]')
        self.assertEqual(len([k for k in gwp.factors()]), 91)


if __name__ == '__main__':
    unittest.main()



from lxml import objectify
from ..file_store import FileStore


class EcoSpoldMasterData(object):

    masters = dict()

    def _read_master(self, filename):
        if filename in self.masters:
            return self.masters[filename]
        else:
            try:
                o = objectify.fromstring(self._fs.readfile(filename + '.xml'))
            except FileNotFoundError:
                o = None
            self.masters[filename] = o
            return o

    __master_units = None
    __master_props = None
    __elem = None
    __inter = None

    @property
    def master_units(self):
        if self.__master_units is None:
            o = self._read_master('Units')
            if o is None:
                self.__master_units = {}
            else:
                self.__master_units = {k.attrib['id']: k for k in o.iterfind('unit', namespaces=o.nsmap)}
        return self.__master_units

    @property
    def master_properties(self):
        if self.__master_props is None:
            o = self._read_master('Properties')
            if o is None:
                self.__master_props = {}
            else:
                self.__master_props = {k.attrib['id']: k for k in o.iterfind('property', namespaces=o.nsmap)}
        return self.__master_props

    @property
    def elementary_exchanges(self):
        if self.__elem is None:
            o = self._read_master('ElementaryExchanges')
            if o is None:
                self.__elem = {}
            else:
                self.__elem = {k.attrib['id']: k for k in o.iterfind('elementaryExchange', namespaces=o.nsmap)}
        return self.__elem

    @property
    def intermediate_exchanges(self):
        if self.__inter is None:
            o = self._read_master('IntermediateExchanges')
            if o is None:
                self.__inter = {}
            else:
                self.__inter = {k.attrib['id']: k for k in o.iterfind('intermediateExchange', namespaces=o.nsmap)}
        return self.__inter

    def __init__(self, source, internal_prefix='MasterData'):
        self._fs = FileStore(source, internal_prefix=internal_prefix)



from collections import defaultdict
import re
import os

from ..archives.term_manager import TermManager, NoFQEntry
from ..contexts import Context, NullContext
from .quelled_cf import QuelledCF
from .clookup import CLookup, SCLookup

from synonym_dict import TermExists, FlowablesDict
# from synonym_dict.example_compartments.compartment_manager import InconsistentLineage  # this is not needed


'''
Switchable biogenic CO2:

* Biogenic CO2 is CO2, so the flowable used to store CFs is 124-38-9 always
* Because flowable is looked up within the quantity implementation, we can assign a synonym to the flow itself, and 
  watch for it
* Then the switch determines whether or not to quell the CF returned to the user, without changing the database
'''
biogenic = re.compile('(biotic|biogenic|non-fossil|in air)', flags=re.IGNORECASE)


DEFAULT_CONTEXTS = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'contexts.json'))
NUM_DEFAULT_CONTEXTS = 36  # added river, long-term

DEFAULT_FLOWABLES = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'flowables.json'))

class QuantityMasqueradeError(Exception):
    pass


class LciaEngine(TermManager):
    """
    This adds several abilities:
     * track flowables by origin
       - identify flowables thhat are not recognized
       - return flowables by origin (all, or only new)
     * lookup CFs based on context hierarchy
       - dist = 0: only exact matchh
       - dist = 1: match or subcompartments
       - dist = 2: match, subcompartments, or parent
       - dist = 3: .. or any parent up to NullContext
     * quell biogenic CO2 in quantity relation lookups
    """
    _quell_biogenic = None

    def _configure_flowables(self, flowables):
        """
        Setup local flowables database with flows that require special handling. Also loads the flowables file.

        When overriding this function, place the super() call between pre-load and post-load activities.
        :return:
        """
        if flowables is None:
            flowables = DEFAULT_FLOWABLES

        # FlowablesDict-- mainly to upsample CAS numbers for matching
        self._fm = FlowablesDict()

        self._fm.new_entry('carbon dioxide', '124-38-9')
        self._fm.new_entry('Water', '7732-18-5')
        # we store the child object and use it to signify biogenic CO2 to optionally quell
        # this strategy depends on the ability to set a query flow's name-- i.e. FlowInterface
        self._bio_co2 = self._fm.new_entry('carbon dioxide (biotic)', '124-38-9', create_child=True)

        # now load + merge from file
        self._fm.load(flowables)

        # now add all known "biotic" synonyms for CO2 to the biotic child
        for k in self._fm.synonyms('124-38-9'):
            if bool(biogenic.search(k)):
                self._bio_co2.add_term(k)
        self._fq_map = {fb: set() for fb in self._fm.objects}

    def __init__(self, contexts=None, flowables=None, quantities=None,
                 quell_biogenic_co2=False,
                 strict_clookup=True,
                 **kwargs):
        """

        :param quantities:
        :param quell_biogenic_co2:
        :param contexts:
        :param flowables:
        :param strict_clookup: [True] whether to prohibit multiple CFs for each quantity / flowable / context tuple
        :param kwargs: from TermManager: quiet, merge_strategy
        """
        if contexts is None:
            contexts = DEFAULT_CONTEXTS
        super(LciaEngine, self).__init__(contexts=contexts, flowables=None, quantities=quantities, **kwargs)

        # override flowables manager
        self._configure_flowables(flowables)

        # the CF lookup: allow hierarchical traversal over compartments [or, um, use a graph db..]
        self._cl_typ = {True: SCLookup,
                        False: CLookup}[strict_clookup]  #
        # another reverse mapping
        self._origins = set()
        self._fb_by_origin = defaultdict(set)  # maps origin to flowables having that origin
        self._fb_by_origin[None] = set(str(k) for k in self._fm.objects)

        # difficult problem, this
        self.quell_biogenic_co2 = quell_biogenic_co2

        self._factors_for_later = defaultdict(bool)

    @property
    def quell_biogenic_co2(self):
        return self._quell_biogenic

    @quell_biogenic_co2.setter
    def quell_biogenic_co2(self, value):
        self._quell_biogenic = bool(value)

    def save_for_later(self, quantity):
        qc = self.get_canonical(quantity)
        if qc is quantity:
            raise QuantityMasqueradeError(quantity)
        self._factors_for_later[qc] = quantity

    def __getitem__(self, item):
        """
        LciaEngine.__getitem__ retrieves a canonical context by more intensively searching for matches from a given
        context.  Adds foreign context's full name as synonym if one is affirmatively found.  If one is not found,
        returns the NullContext.

        None is returned as None, to represent 'unspecified' (i.e. accept all) as opposed to 'no context' which isa
        context (accept only matching). (as tested)

        :param item:
        :return: a matching context or NullContext.
        """
        if item is None:
            return None
        try:
            return self._cm.__getitem__(item)
        except KeyError:
            if isinstance(item, Context):
                return self._cm.find_matching_context(item)
            elif isinstance(item, tuple) and len(item) > 1:
                return self.__getitem__(item[1:])
            return NullContext

    def apply_hints(self, names, hints):
        """
        Hints should be
        :param names: iterable of catalog names to which context hints will apply
        :param hints: an iterable of term, canonical pairs, where term is the context as known in the origin, and
        canonical is the corresponding canonical context.
        :return:
        """
        orgs = list(names)
        for hint_type, term, canonical in hints:
            if term in self.synonyms(canonical):
                continue
            if hint_type == 'context':
                for org in orgs:
                    print('Applying context hint %s:%s => %s' % (org, term, canonical))
                    self._cm.add_context_hint(org, term, canonical)
            elif hint_type == 'quantity':
                print('Applying quantity hint %s -> %s' % (term, canonical))
                try:
                    self._qm.add_synonym(canonical, term)
                except TermExists:
                    assert self._qm[canonical] is self._qm[term]
            elif hint_type == 'flowable':
                print('Applying flowable hint %s -> %s' % (term, canonical))
                try:
                    self._fm.add_synonym(canonical, term)
                except TermExists:
                    assert self._fm[canonical] == self._fm[term]
            else:
                raise ValueError('Unknown hint type %s' % hint_type)

    def add_synonym(self, existing_term, synonym):
        """

        :param existing_term: flowable or quantity (cx is TBD)
        :param synonym: string term to add
        :return:
        """
        synonym = str(synonym).strip()
        try:
            ent = self._fm[existing_term]
            self._add_to_existing_flowable(ent, (synonym,))
        except KeyError:
            try:
                ent = self._qm[existing_term]
                self._qm.add_synonym(ent, synonym)
            except KeyError:
                raise KeyError('No entry found for %s' % existing_term)

    @staticmethod
    def _flow_terms(flow):
        """
        This function was created because we don't want TermManagers getting confused about things like misassigned
        CAS numbers, because they don't have the capacity to store multiple CFs.  So we will save the full synonym
        list for the LciaEngine.
        :param flow:
        :return:
        """
        return flow.synonyms

    def _add_to_existing_flowable(self, fb, new_terms):
        """
        Harvest biogenic co2 synonyms (not worth correcting for other biogenic substances??)
        :param fb:
        :param new_terms:
        :return:
        """
        biog = ('124-38-9' in fb)
        for term in new_terms:
            self._fm.add_synonym(fb, term)
            if biog and bool(biogenic.search(term)):
                self._bio_co2.add_term(term)  # ensure that bio term is a biogenic synonym

    def _add_flow_terms(self, flow, merge_strategy=None):
        """
        Subclass handles two problems: tracking flowables by origin and biogenic CO2.

        Should probably test this shit

        Under our scheme, it is a requirement that the flowables list used to initialize the LciaEngine is curated.

        biogenic: if ANY of the flow's terms match the biogenic regex AND the flow is CO2, set its name
        :param flow:
        :return:
        """
        fb = super(LciaEngine, self)._add_flow_terms(flow, merge_strategy=merge_strategy)
        self._fb_by_origin[flow.origin].add(str(fb))
        if '124-38-9' in fb:
            try:
                bio = next(t for t in flow.synonyms if bool(biogenic.search(t)))
            except StopIteration:
                # no biogenic terms
                return fb
            flow.name = bio  # ensure that flow's name shows up with that term
            self._fb_by_origin[flow.origin].add(bio)
        return fb

    def add_quantity(self, quantity):
        """
        Here we are not picky about having duplicate quantities
        :param quantity:
        :return:
        """
        if quantity.entity_type != 'quantity':
            raise TypeError('Must be quantity type')
        if quantity.link not in self._qm:
            self._qm.add_quantity(quantity)
            assert quantity.link in self._qm
        return self._canonical_q(quantity)

    '''
    def get_canonical(self, quantity):
        """
        Override TermManager default to automatically add unknown quantities
        :param quantity:
        :return:
        """
        try:
            return self._canonical_q(quantity)
        except KeyError:
            if hasattr(quantity, 'entity_type') and quantity.entity_type == 'quantity' and not quantity.is_entity:
                print('Missing canonical quantity-- adding to LciaDb')
                self._catalog.register_quantity_ref(quantity)
                q_can = self._tm.get_canonical(quantity)
    '''

    def merge_quantities(self, first, second):
        """
        Absorb second into child of first.  Currently does not support remapping entries, so import_cfs(second)
        needs to be run again. Old factors will be left in so _fq_map still works.
        :param first:
        :param second:
        :return:
        """
        dom = self.get_canonical(first)
        add = self.get_canonical(second)
        self._qm.merge(dom, add)
        self.import_cfs(second)

    def _add_compartments(self, comps):
        """
        For LciaEngines, we use the default behavior of ContextManager which is to first try 'attach', then fallback
        to 'rename'
        :param comps:
        :return:
        """
        return self._cm.add_compartments(comps)

    '''  # This doesn't do anything useful
    def add_subcontext(self, context, prefix, sub, origin=None):
        """

        :param context: a canonical context
        :param prefix:
        :param sub: a new subcompartment
        :param origin: [None]
        :return:
        """
        if context is NullContext or context is None or context == ():
            return context
        cx_sub = tuple(context.as_list() + ['%s-%s' % (prefix, sub)])

        try:
            cx = self._cm[cx_sub]
        except (KeyError, InconsistentLineage):
            cx = self.add_context(cx_sub, origin=origin)
        return cx
    '''

    def import_cfs(self, quantity):
        """
        Given a quantity, import its CFs into the local database.  Unfortunately this is still going to be slow because
        every part of the CF still needs to be canonicalized. The only thing that's saved is creating a new
        Characterization instance.
        :param quantity: this is a TRUE quantity whose query points to the authentic origin.
        :return:
        """
        try:
            qq = self._canonical_q(quantity)
        except KeyError:
            qq = self.add_quantity(quantity)

        count = 0
        for cf in quantity.factors():
            count += 1
            # print(cf)
            try:
                fb = self._fm[cf.flowable]
            except KeyError:
                fb = self._create_flowable(*quantity.query_synonyms(cf.flowable))

            self.add_quantity(cf.ref_quantity)  # this may lead to the creation of non-converting quantities if units mismatch

            cx = self[cf.context]

            self._qassign(qq, fb, cf, context=cx)
        self._factors_for_later[quantity] = True
        print('Imported %d factors for %s' % (count, quantity))

    def _check_factors(self, qq):
        if hasattr(self._factors_for_later[qq], 'factors'):
            self.import_cfs(self._factors_for_later.pop(qq))

    def _find_exact_cf(self, qq, fb, cx, origin, flowable):
        try:
            ql = self._qlookup(qq, fb)
        except NoFQEntry:
            return None
        # cfs = ql._context_origin(cx, origin=origin)
        # if fb is self._fm['water']:  # this is not how to handle this
        #     cx = self.add_subcontext(cx, 'flow', flowable)
        cfs = ql.find(cx, dist=0, origin=origin)  # sliiiiightly slower but much better readability
        if len(cfs) > 0:
            if len(cfs) > 1:  # can only happen if qq's origin is None ??
                try:
                    return next(cf for cf in cfs if cf.origin is None)
                except StopIteration:
                    # this should never happen
                    return next(cf for cf in cfs if cf.origin == qq.origin)
            return list(cfs)[0]
        return None

    @staticmethod
    def _store_cf(cl, context, new_cf):
        """
        Assigns the cf to the mapping; does subclass-specific collision checking
        :param cl:
        :param new_cf:
        :return:
        """
        try:
            cl.add(new_cf, key=context)
        except TypeError:
            print(type(cl))
            raise

    def _qassign(self, qq, fb, new_cf, context=None):
        """

        :param qq:
        :param fb:
        :param new_cf:
        :param context:
        :return:
        """
        ''' # this is done in _find_exact_cf already
        # don't characterize "Water" per se-- give it a subcontext
        if context is not NullContext and fb is self._fm['Water']:
            context = self.add_subcontext(context, 'flow', new_cf.flowable)
        '''

        super(LciaEngine, self)._qassign(qq, fb, new_cf, context)
        self._origins.add(new_cf.origin)

    def merge_flowables(self, dominant, *syns):
        for syn in syns:
            self._fm.merge(dominant, syn)

    def save_flowables(self, filename=None):
        self._fm.save(filename)

    def save_contexts(self, filename=None):
        self._cm.save(filename)

    def flowables(self, search=None, origin=None, new=False, **kwargs):
        """
        Adds ability to filter by origin-- note this is exclusive to the ability to filter by quantity
        :param search:
        :param origin:
        :param new: [False] if True, only return flowables that were not known originally
        :return: generates flowable strings
        """
        if origin is None:
            _iter = super(LciaEngine, self).flowables(search=search, **kwargs)
        else:
            if search is None:
                _iter = self._fb_by_origin[origin]
            else:
                _iter = (str(x) for x in self._fm.objects_with_string(search) if x in self._fb_by_origin[origin])

        for k in _iter:
            if new:
                if k in self._fb_by_origin[None]:
                    continue
            yield k

    @staticmethod
    def is_biogenic(term):
        return bool(biogenic.search(term))

    def _quell_co2(self, flowable, context):
        """
        We assume that all biogenic CO2 flows will be detected via add_flow, and will have their names set to something
        known to our _bio_co2 Flowable. So: If we are quelling, and if the flowable (string) is synonym to _bio_co2,
        we ask if it is a resource from air, or if it is any emission.
        :param flowable: orig, not looked-up in _fm
        :param context: from CF
        :return: bool
        """
        if self._quell_biogenic is False:
            return False
        if flowable in self._bio_co2:
            if context.is_subcompartment(self._cm['from air']):
                return True
            if context.is_subcompartment(self._cm['Emissions']):
                return True
        return False

    def _factors_for_flowable(self, fb, qq, cx, **kwargs):
        """
        detach lookup for cleanness. canonical everything
        :param fb:
        :param qq:
        :param cx:
        :param kwargs: used in subclasses
        :return:
        """
        self._check_factors(qq)
        try:
            cl = self._qlookup(qq, fb)
        except NoFQEntry:
            return
        if cx is None:
            for v in cl.cfs():
                yield v
        else:
            for v in cl.find(cx, **kwargs):
                yield v

    def factors_for_flowable(self, flowable, quantity=None, context=None, **kwargs):
        """
        Here we deal with water contexts specified as flowables by creating children- this is generalizeable (vs for CO2)
        :param flowable:
        :param quantity:
        :param context:
        :param kwargs:
        :return:
        """
        try:
            fb = self._fm[flowable]
        except KeyError:
            return
        '''  # this does not do anything helpful
        if fb == self._fm['Water']:
            try:
                context = self._cm['flow-%s' % flowable]
            except KeyError:
                pass
        '''
        for k in super(LciaEngine, self).factors_for_flowable(flowable, quantity=quantity, context=context, **kwargs):
            if self._quell_co2(flowable, context):
                yield QuelledCF.from_cf(k, flowable=self._bio_co2)
            else:
                yield k

    def factors_for_quantity(self, quantity, flowable=None, context=None, **kwargs):
        self._check_factors(self._canonical_q(quantity))
        return super(LciaEngine, self).factors_for_quantity(quantity, flowable=flowable, context=context, **kwargs)


    def _serialize_qdict(self, origin, quantity, values=False):
        _ql = self._qaccess(quantity)
        d = {}
        for fb, cl in _ql.items():
            _od = cl.serialize_for_origin(origin, values=values)
            if len(_od) > 0:
                d[str(fb)] = _od
        return d



"""
Change Log
2020-01-21: Bump count of flowables to 4005 bc of Ecoinvent synonym fix
"""

import unittest

from .. import LciaDb
from ..lcia_engine import NUM_DEFAULT_CONTEXTS
from ...entities import LcQuantity, LcFlow
from ...contexts import Context
from ...archives.tests.test_qdb import num_ref_quantities

class LciaEngineTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.lcia = LciaDb.new()

    def test_0_init(self):
        self.assertEqual(len([x for x in self.lcia.query.contexts()]), NUM_DEFAULT_CONTEXTS + 1)  # add Intermediate Flows
        self.assertEqual(len([x for x in self.lcia.query.flowables()]), 4005)
        self.assertEqual(len([x for x in self.lcia.query.quantities()]), num_ref_quantities)
        self.assertEqual(len(self.lcia.tm._q_dict), 3)

    def test_1_add_characterization(self):
        rq = self.lcia.query.get_canonical('mass')
        qq = self.lcia.query.get_canonical('volume')

        cf = self.lcia.query.characterize('water', rq, qq, .001)
        self.assertEqual(cf.origin, self.lcia.ref)

    def test_2_cf(self):
        self.assertEqual(self.lcia.query.quantity_relation('water', 'mass', 'volume', None).value, .001)
        self.assertEqual(self.lcia.query.quantity_relation('water', 'volume', 'mass', None).value, 1000.0)

    def test_3_dup_mass(self):
        dummy = 'dummy_external_ref'
        dup_mass = LcQuantity(dummy, referenceUnit='kg', Name='Mass', origin='dummy.origin')
        self.lcia.tm.add_quantity(dup_mass)
        self.assertEqual(self.lcia.query.get_canonical(dummy), self.lcia.query.get_canonical('mass'))

    def test_add_flow(self):
        """
        In this case, the flow's synonyms should be added. But dummy flows' links are not added to synonym set.
        :return:
        """
        flow = LcFlow('Dummy Flow', ref_qty=self.lcia.tm.get_canonical('mass'), origin='test.origin')
        for k in ('phosphene', 'phxphn', '1234567'):
            flow._flowable.add_term(k)
        self.lcia.tm.add_flow(flow)
        self.assertEqual(self.lcia.tm._fm['1234-56-7'].name, 'test.origin/Dummy Flow')

    def test_add_non_matching_context(self):
        """
        This test is actually really thorny because 'unspecified', on its own, returns NullContext as a forbidden
        term, which is different from 'elementary flows', which raises a KeyError.
        :return:
        """
        cx = self.lcia.tm['emissions to air']
        c0 = Context('Elementary flows')
        c1 = Context('Emission to air', parent=c0)
        c2 = Context('unspecified', parent=c1)
        c2.add_origin('test')
        self.assertIs(self.lcia.tm[c2], cx)

    def test_add_conflicting_contexts(self):
        c1 = self.lcia.tm.add_context(('emissions', 'emissions to water'))
        c2 = self.lcia.tm.add_context(('elementary flows', 'emissions to water'))
        self.assertIs(c1, c2)
        self.assertEqual(len(c1.seq), 3)


if __name__ == '__main__':
    unittest.main()



try:
    import xlrd
except ImportError:
    xlrd = False

try:
    import openpyxl
except ImportError:
    openpyxl = False


class XlDict(object):
    """
    wrapper class for xlrd that exposes a simple pandas-like interface to access tabular spreadsheet data with iterrows.
    """
    @classmethod
    def from_sheetname(cls, workbook, sheetname, **kwargs):
        return cls(workbook.sheet_by_name(sheetname), **kwargs)

    def __init__(self, sheet, nulls=None):
        """

        :param sheet: an xlrd.sheet.Sheet
        :param nulls: What value to assign to empty entries.  Default is to set them to None.  Pass in the desired
        literal
        """
        self._sheet = sheet
        self._nulls = nulls

    def _map_value(self, value):
        if not isinstance(value, str):
            return value
        value = value.strip()
        if value == '':
            return self.nulls
        return value

    @property
    def nulls(self):
        return self._nulls

    @nulls.setter
    def nulls(self, value):
        self._nulls = value

    def _row_gen(self):
        """
        A method that generates rows as iterables
        :return:
        """
        return self._sheet.get_rows()

    def iterrows(self):
        """
        Using the first row as a list of headers, yields a dict for each subsequent row using the header names as keys.
        returning index, row for pandas compatibility
        :return:
        """
        _gen = self._row_gen()
        # grab first row
        d = dict((v.value, k) for k, v in enumerate(next(_gen)))
        index = 0
        for r in _gen:
            index += 1
            yield index, dict((k, self._map_value(r[v].value)) for k, v in d.items())

    def unique_units(self, internal=False):
        """
                unitname = 'unit' if self.internal else 'unitName'
        units = set(_elementary[unitname].unique().tolist()).union(
            set(_intermediate[unitname].unique().tolist()))
        for u in units:
            self._create_quantity(u)

        :param internal:
        :return:
        """
        units = set()
        unitname = 'unit' if internal else 'unitName'
        for index, row in self.iterrows():
            units.add(row[unitname])
        return units


class XlsxDict(XlDict):
    """
    Compatibility adapter to use openpyxl spreadsheets instead

    """
    @classmethod
    def from_sheetname(cls, workbook, sheetname, **kwargs):
        return cls(workbook[sheetname], **kwargs)

    def _row_gen(self):
        return self._sheet.iter_rows()


def xl_dict(file, sheetname, **kwargs):
    f = []
    if xlrd:
        try:
            wb = xlrd.open_workbook(file)
            return XlDict.from_sheetname(wb, sheetname, **kwargs)
        except AttributeError:
            f.append('xlrd')
    if openpyxl:
        try:
            wb = openpyxl.load_workbook(file)
            return XlsxDict.from_sheetname(wb, sheetname, **kwargs)
        except TypeError:
            f.append('openpyxl')
    if len(f) == 0:
        raise ImportError('Unable to import any excel file readers')
    else:
        raise TypeError('%s failed to open file' % ', '.join(f))




import unittest
from .base_testclass import BasicEntityTest


class ProcessesTest(BasicEntityTest):
    """
    Things to test:
    """
    def test_reference(self):
        diesel = next(self.A.search('flow', Name='diesel, at refinery'))
        self.assertAlmostEqual(self.petro.reference(diesel).value, 0.000252345)

    def test_single_reference(self):
        """
        For a process with exactly one reference, it should be returned without specification
        :return:
        """
        elec = next(self.A.search('flow', Name='Electricity, at grid, US, 2008'))
        ex = self.grid.reference()
        self.assertIs(elec, ex.flow)

    def test_inventory(self):
        """
        inventory() returns unallocated exchanges and inventory(ref) returns exchanges normalized by ref
        :return:
        """
        ex = self.grid.reference()
        inv = self.grid.inventory()
        inv_dict = {x.flow: x.value for x in inv}
        for x in self.grid.inventory(ex):
            self.assertEqual(x.value, inv_dict[x.flow] / ex.value)

    def test_allocation_factors(self):
        af = self.petro.allocation_factors()
        self.assertEqual(len(af), 9)
        self.assertAlmostEqual(sum(af.values()), 1.0, places=12)

    def test_inv_not_reference(self):
        """
        test implementation of the interface specification for inventory

        exchange-finding tools have been fixed but the test has yet to be written
        :return:
        """
        pass


if __name__ == '__main__':
    unittest.main()



"""
Contexts in this sense are environmental compartments, except they have the added capability to keep lists of origins.

Edelen and Ingwersen et al 2017:
"Recommendations: ...setting an exclusive or inclusive nomenclature for flow context information that includes
directionality and environmental compartment information."

In the antelope architecture, there are two different objectives for handling contexts as-presented by the data source.

 In the default case, for every static resource or stand-alone archive a "TermManager" is created which is captive to
 the archive.  The role of this class is to collect information from the data source in as close to its native
 presentation as possible. This creates an "inclusive" nomenclature for the source.

 In the Catalog case, catalog's local quantity DB is an LciaEngine, which is also shared among all foreground
 resources.  In this case the objective is to match a given context to the existing (exclusive) nomenclature built-in
 to the LciaEngine, so that contexts are guaranteed to coincide during LCIA.

In order to accomplish this, the native add_context() method needs to be expansive, fault tolerant, and widely accepting
of diverse inputs, whereas find_matching_context() needs to be more discerning and rigorous.  Thus the former accepts
a tuple of string terms or a context, but the latter requires a context.  find_matching_context searches first bottom-up
then top-down to look for matches.  It requires an input context to have an origin already specified, so that the
resources can assign "hints" that map local terms to canonical contexts on an origin-specific basis.

Note that a "matching" generally requires an exact (case-insensitive) match to an entry in the contexts database.
Curation of the synonym set is required.

Because contexts must be directional, some terms are protected as ambiguous: "air", "water", and "ground" should be
avoided in favor of explicit "from air" or "to air" or synonyms.

The NullContext is a singleton defined in this file that is meant to imply no specific context.  It is interpreted in
two different ways:
 - on the characterization side, NullContext indicates the characterization has no specific context, and thus applies to
   all contexts, as long as a more applicable characterization is not found.
 - on the query side, NullContext indicates that a context was specified but no match was found.  Queries with
   NullContext should not match any existing characterizations (except NullContext itself).

The NullContext should be returned by the context manager
"""

from synonym_dict import Compartment, CompartmentManager, NonSpecificCompartment
from synonym_dict.compartments.compartment import InvalidSubCompartment
from antelope import valid_sense

ELEMENTARY = {'elementary flows', 'resource', 'emission', 'resources', 'emissions'}

PROTECTED = ('air', 'water', 'ground')


class ProtectedTerm(Exception):
    """
    currently triggered by the stale contexts file-- leading to an error in which
    "emissions;to water;low population density, long-term" and
    "emissions;to air;low population density, long-term"
    are synonyms in basic TermManagers. Should not present problems in LciaEngines, but the specific case is not tested.

    Solution is to patch the local-contexts.json file to exclude this redundant term for water.
    Justification: the canonical contexts should be well-behaved and not contain any inherent conflicts. It is
    a nomenclature problem for the two contexts to have the same name.
    """
    pass


class InconsistentSense(Exception):
    pass


class FrozenElementary(Exception):
    """
    top-level elementary contexts may not be assigned non-elementary parents
    """
    pass


def _dir_mod(arg, sense):
    if arg.lower() in PROTECTED:
        if sense is None:
            raise ProtectedTerm(arg)
        mod = {'Source': 'from ', 'Sink': 'to '}[sense]
        arg = '%s%s' % (mod, arg)
    return arg


class Context(Compartment):
    """
    If 'resources' or 'emissions' match any terms in a compartment, it is considered 'elementary', along with all its
    subcompartments.

    A context has a natural directional "sense", which is either 'Source', 'Sink', or None.  A Source context
    generates flows which may be inputs to the activity; a Sink context absorbs flows which are output from the
    activity.

    If a context has a parent, it inherits the sense of the parent- specifying the opposite sense will raise
    an error.
    """
    _first_origin = None
    entity_type = 'context'
    _elem = None

    @property
    def origin(self):
        return self._first_origin

    @staticmethod
    def validate():
        """
        Need this for termination checking
        :return:
        """
        return True

    @property
    def fullname(self):
        if self._first_origin is None:
            return self.name
        return '%s:%s' % (self.origin, self.name)

    @property
    def origins(self):
        for o in self._origins:
            yield o

    def __init__(self, *args, sense=None, **kwargs):
        self._sense = None
        super(Context, self).__init__(*args, **kwargs)
        self._origins = set()
        if sense is not None:
            self.sense = sense

    @property
    def terms(self):
        if self._first_origin is not None:
            yield self.fullname
        for t in super(Context, self).terms:
            yield t

    @property
    def sense(self):
        if self._sense is not None:
            return self._sense
        elif self.parent is None:
            return self._sense
        return self.parent.sense

    def _check_sense(self, v_sense):
        if self.sense is not None and self.sense != v_sense:
            raise InconsistentSense('Value %s conflicts with current v_sense %s' % (v_sense, self.sense))
        for sub in self.subcompartments:
            sub._check_sense(v_sense)

    @sense.setter
    def sense(self, value):
        sense = valid_sense(value)
        self._check_sense(sense)
        self._sense = sense
        #for sub in self.subcompartments:  # do not set subcompartment senses- only set determinative compartments
        #    sub.sense = sense

    @property
    def parent(self):  # duplicating here to override setter
        return self._parent

    @parent.setter
    def parent(self, parent):
        if self._parent is None:
            if parent:
                if self.elementary and not parent.elementary:
                    raise FrozenElementary
            self._elem = None  # reset
        if parent:
            if parent.sense is not None:
                self._check_sense(parent.sense)
        # print('Setting %s.parent <- %s' % (self, parent))
        if str(self) in PROTECTED and str(parent) in PROTECTED and str(self) != str(parent):
            raise ProtectedTerm
        Compartment.parent.fset(self, parent)

    def _is_elem(self):
        for t in self.terms:
            if t.strip().lower() in ELEMENTARY:
                return True
        return False

    @property
    def elementary(self):
        """
        A context's "elementary-ness" is determined by the top-most parent.  Everything inherits from above.
        If the (lower cased) parent context is found in ELEMENTARY, it's elementary.

        The only way to make a context elementary is to name it such, or give it an elementary parent.
        :return:
        """
        if self.parent is None:
            if self._elem is None:
                self._elem = self._is_elem()
            return self._elem
        return self.parent.elementary

    @property
    def external_ref(self):
        return self.name

    def add_origin(self, origin):
        if self._first_origin is None:
            self._first_origin = origin
        self._origins.add(origin)
        if self.parent is not None:  # why would we recurse on this?
            # ans: if our parent is also missing an origin
            if self.parent.origin is None:
                self.parent.add_origin(origin)

    def has_origin(self, origin, strict=False):
        try:
            if strict:
                next(x for x in self._origins if x == origin)
            else:
                next(x for x in self._origins if x.startswith(origin))
        except StopIteration:
            return False
        return True

    def serialize(self):
        d = super(Context, self).serialize()
        if self.parent is None:
            if self.sense is not None:
                d['sense'] = self.sense
        return d

    def __repr__(self):
        return '<Context(%s)>' % ';'.join(self.as_list())


NullContext = Context.null()


class ContextManager(CompartmentManager):
    _entry_group = 'Compartments'  # we keep this so as to access compartment-compatible serializations
    _syn_type = Context
    _ignore_case = True

    _null_entry = NullContext

    def __init__(self, source_file=None):
        super(ContextManager, self).__init__()

        self.new_entry('Resources', sense='source')
        self.new_entry('Emissions', sense='sink')
        self.load(source_file)

    def add_context_hint(self, origin, term, canonical):
        """
        Method for linking foreign context names to canonical contexts
        :param origin: origin of foreign context
        :param term: foreign context name
        :param canonical: recognized synonym for canonical context that should match the foreign one
        :return:
        """
        c = self[canonical]
        if c is None:
            raise ValueError('Unrecognized canonical context %s' % c)
        syn = '%s:%s' % (origin, term)
        self.add_synonym(c, syn)

    def new_entry(self, *args, parent=None, **kwargs):
        args = tuple(filter(None, args))
        if parent is not None:
            if not isinstance(parent, Compartment):
                parent = self._d[parent]
            if parent.sense is not None:
                args = tuple(_dir_mod(arg, parent.sense) for arg in args)
        return super(ContextManager, self).new_entry(*args, parent=parent, **kwargs)

    def _gen_matching_entries(self, cx, sense):
        for t in cx.terms:
            mt = _dir_mod(t, sense)
            if mt in self._d:
                yield self._d[mt]

    def _merge(self, existing_entry, ent):
        """
        Need to check lineage. We adopt the rule: merge is acceptable if both entries have the same top-level
        compartment or if ent has no parent.  existing entry will obviously be dominant, but any sense specification
        will overrule a 'None' sense in either existing or new.
        :param existing_entry:
        :param ent:
        :return:
        """
        if ent.sense is not None:
            existing_entry.sense = ent.sense  # this is essentially an assert w/raises InconsistentSense

        super(ContextManager, self)._merge(existing_entry, ent)  # yowza, moronic bug

    def add_compartments(self, comps, conflict=None):
        if conflict is not None:
            return super(ContextManager, self).add_compartments(comps, conflict=conflict)
        else:
            try:
                return super(ContextManager, self).add_compartments(comps, conflict='attach')
            except (FrozenElementary, InvalidSubCompartment, InconsistentSense):
                return super(ContextManager, self).add_compartments(comps, conflict='rename')
            """
            except ProtectedTerm:
                print('Protected Term! %s' % comps)
                return super(ContextManager, self).add_compartments(comps, conflict='rename')
            """

    '''
    def add_lineage(self, lineage, parent=None):
        """
        Create a set of local contexts, beginning with parent, that replicate those in lineage
        :param lineage:
        :param parent: [None] parent of lineage
        :return: the last created context
        """
        if parent:
            if parent not in self._l:
                raise ValueError('Parent is not known locally')
        new = None
        for lx in lineage:
            new = self.new_entry(*lx.terms, parent=parent)
            if lx.origin is not None:
                self.add_synonym(lx.fullname, new)
            parent = new
        return new
    '''

    def _add_but_not_protected(self, entry, context):
        self.add_synonym(entry, context.fullname)
        if context.name.lower() not in PROTECTED:
            self.add_synonym(entry, context.name)

    def find_matching_context(self, context):
        if context.name == context.fullname:
            raise AttributeError('Context origin must be specified')
        current = NullContext  # current = deepest local match

        # first, look for stored auto_names or context_hints to find an anchor point:
        missing = []
        for cx in context.seq[::-1]:
            if cx.fullname in self:
                current = self[cx.fullname]
                break
            else:
                missing = [cx] + missing  # prepend

        # if current is None when we get here, then we haven't found anything, so start over and hunt from bottom up
        while len(missing) > 0:
            this = missing.pop(0)  # this = active foreign match
            if current is NullContext:
                try:
                    current = next(self._gen_matching_entries(this, None))
                except StopIteration:
                    continue
                except ProtectedTerm:
                    continue
                if current is NullContext:
                    continue
                self._add_but_not_protected(current, this)
            else:
                try:
                    nxt = next(k for k in self._gen_matching_entries(this, current.sense)
                               if k.is_subcompartment(current))
                    self._add_but_not_protected(nxt, this)
                    current = nxt
                except StopIteration:
                    self._add_but_not_protected(current, this)
        if current is not NullContext:
            self._add_but_not_protected(current, context)
        return current

    def __getitem__(self, item):
        try:
            if isinstance(item, Context):
                try:
                    return super(ContextManager, self).__getitem__(item.fullname)
                except (NonSpecificCompartment, KeyError):
                    pass
            return super(ContextManager, self).__getitem__(item)
        except NonSpecificCompartment:
            return NullContext



from collections import namedtuple


QInfo = namedtuple('QInfo', ['Category', 'Compartment', 'RefUnit'])


ns_uuid_21 = '150e35c3-ac4a-485b-826b-a41807ddf43a'

q_info_21 = {
    'Global Warming Air (kg CO2 eq / kg substance)': QInfo('Global Warming Air',  'air', 'kg CO2 eq'),
    'Acidification Air (kg SO2 eq / kg substance)': QInfo('Acidification Air', 'air',  'kg SO2 eq'),
    'HH Particulate Air (PM2.5 eq / kg substance)': QInfo('Human Health Particulates Air', 'air', 'PM2.5 eq'),
    'Eutrophication Air (kg N eq / kg substance)': QInfo('Eutrophication Air', 'air', 'kg N eq'),
    'Eutrophication Water (kg N eq / kg substance)': QInfo('Eutrophication Water', 'water', 'kg N eq'),
    'Ozone Depletion Air (kg CFC-11 eq / kg substance)': QInfo('Ozone Depletion Air', 'air', 'kg CFC-11 eq'),
    'Smog Air (kg O3 eq / kg substance)': QInfo('Smog Air', 'air', 'kg O3 eq'),
    'Ecotox. CF [CTUeco/kg], Em.airU, freshwater': QInfo('Ecotoxicity, freshwater', 'urban air', 'CTUeco'),
    'Ecotox. CF [CTUeco/kg], Em.airC, freshwater': QInfo('Ecotoxicity, freshwater', 'rural air', 'CTUeco'),
    'Ecotox. CF [CTUeco/kg], Em.fr.waterC, freshwater': QInfo('Ecotoxicity, freshwater', 'fresh water', 'CTUeco'),
    'Ecotox. CF [CTUeco/kg], Em.sea waterC, freshwater': QInfo('Ecotoxicity, freshwater', 'sea water', 'CTUeco'),
    'Ecotox. CF [CTUeco/kg], Em.nat.soilC, freshwater': QInfo('Ecotoxicity, freshwater', 'soil', 'CTUeco'),
    'Ecotox. CF [CTUeco/kg], Em.agr.soilC, freshwater': QInfo('Ecotoxicity, freshwater', 'agricultural', 'CTUeco'),
    'Human health CF  [CTUcancer/kg], Emission to urban air, cancer': QInfo(
        'Human health toxicity, cancer', 'urban air', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to urban air, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'urban air', 'CTUnoncancer'),
    'Human health CF  [CTUcancer/kg], Emission to cont. rural air, cancer': QInfo(
        'Human health toxicity, cancer', 'rural air', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to cont. rural air, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'rural air', 'CTUnoncancer'),
    'Human health CF  [CTUcancer/kg], Emission to cont. freshwater, cancer': QInfo(
        'Human health toxicity, cancer', 'fresh water', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to cont. freshwater, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'fresh water', 'CTUnoncancer'),
    'Human health CF  [CTUcancer/kg], Emission to cont. sea water, cancer': QInfo(
        'Human health toxicity, cancer', 'sea water', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to cont. sea water, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'sea water', 'CTUnoncancer'),
    'Human health CF  [CTUcancer/kg], Emission to cont. natural soil, cancer': QInfo(
        'Human health toxicity, cancer', 'soil', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to cont. natural soil, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'soil', 'CTUnoncancer'),
    'Human health CF  [CTUcancer/kg], Emission to cont. agric. Soil, cancer': QInfo(
        'Human health toxicity, cancer', 'agricultural', 'CTUcancer'),
    'Human health CF  [CTUnoncancer/kg], Emission to cont. agric. Soil, non-canc.': QInfo(
        'Human health toxicity, non-cancer', 'agricultural', 'CTUnoncancer')
}



from ..clookup import CLookup, SCLookup, FactorCollision, QuantityMismatch, DuplicateOrigin, Context
from ...entities import LcFlow, LcQuantity
from ...characterizations import Characterization
import unittest


q1 = LcQuantity.new('A quantity', 'kg')
q2 = LcQuantity.new('Another quantity', 'MJ ncv')
q3 = LcQuantity.new('A third quantity', 'MJ gcv')

f1 = LcFlow.new('Massive flow', q1, Compartment=['emissions', 'to air'])
f1d = LcFlow.new('Massive flow', q1, Compartment=['emissions', 'to air'])
f2 = LcFlow.new('Energetic flow', q2, Compartment=['resources', 'from ground'])

cx_air = Context('to air')
cx_ua = Context('to urban air', parent=cx_air)
cx_ra = Context('to rural air', parent=cx_air)
cx_r = Context('resources')
cx_rg = Context('from ground', parent=cx_r)

cf = Characterization.from_flow(f1, q2, context=cx_air, value=47)
cfo = Characterization.from_flow(f1d, q2, context=cx_air, value=92, origin='dummy')
cfd = Characterization.from_flow(f1d, q2, context=cx_air, value=84)
cfua = Characterization.from_flow(f1, q2, context=cx_ua, value=58)
cfra = Characterization.from_flow(f1, q2, context=cx_ra, value=51)

cg = Characterization.from_flow(f2, q2, context=cx_r, value=123)
ch = Characterization.from_flow(f2, q3, context=Context.null(), value=1.023)


class SingleCfTest(unittest.TestCase):
    def test_create_clookup(self):
        g = CLookup()
        g.add(cf)
        self.assertIs(g.find_first(cx_air, dist=0), cf)

    def test_strict(self):
        g = CLookup()
        gs = SCLookup()
        g.add(cf)
        gs.add(cf)
        ''' # now we are allowing duplicate origins
        with self.assertRaises(DuplicateOrigin):
            g.add(cfd)
        '''
        with self.assertRaises(FactorCollision):
            gs.add(cfd)
        g.add(cfd)
        g.add(cfo)
        with self.assertRaises(FactorCollision):
            gs.add(cfo)

    def test_quantity_mismatch(self):
        g = CLookup()
        g.add(cg)
        with self.assertRaises(QuantityMismatch):
            g.add(ch)

    def test_find_1(self):
        g = CLookup()
        g.add(cfua)
        g.add(cfra)
        self.assertEqual(len(g.find(cx_air, dist=0)), 0)
        self.assertSetEqual({x.value for x in g.find(cx_air, dist=1, return_first=False)}, {51, 58})

    def test_find_2(self):
        g = CLookup()
        g.add(cg)
        self.assertEqual(len(g.find(cx_rg, dist=0)), 0)
        self.assertEqual(len(g.find(cx_rg, dist=1)), 0)
        self.assertListEqual(g.find(cx_rg, dist=2), [cg])


if __name__ == '__main__':
    unittest.main()



import os

from .ilcd import grab_flow_name
from ..xml_widgets import *
from .ilcd_lcia import IlcdLcia

ELCD = os.path.join('/data', 'LCI', 'ELCD', 'ELCD3.2.zip')


def ilcd_flow_generator(archive=ELCD, **kwargs):
    """
    This generates flows from the current reference ELCD archive.
    :param archive:
    :param kwargs:
    :return:
    """
    i = IlcdLcia(archive, **kwargs)
    count = 0
    for f in i.list_objects('Flow'):
        o = i.objectify(f, dtype='Flow')
        if o is not None:
            yield o
            count += 1
            if count % 1000 == 0:
                print('%d data sets completed' % count)


ilcd_bad_synonyms = {
    'fe0acd60-3ddc-11dd-a6f8-0050c2490048': "Crude Oil; 42.3 MJ/kg",
    '08a91e70-3ddc-11dd-944a-0050c2490048': "Carbon, resource, in ground",
    '3e4d9eab-6556-11dd-ad8b-0800200c9a66': "Wood;  14.7 MJ/kg"
}


ilcd_bad_cas = {

}


def synonyms_from_ilcd_flow(flow, skip_syns=False):
    """
    ILCD flow files have long synonym blocks at the top. They also have a CAS number and a basename.

    Skips synonym blocks for ILCD flows known to have bad synonyms:
      * "Crude Oil; 42.3 MJ/kg" is not a synonym for "Benzene, pure", etc.
      * "Carbon [resource, in ground]" is not a synonym for the variety of compounds that may be manufactured from it
      * "Wood;  14.7 MJ/kg" says synonyms removed but weren't.

    Skips 'wood' from any list, which is abused badly in the ILCD synonyms list. Methanol and wood are not synonymous.
    :param flow:
    :param skip_syns: return name, uuid, and cas but skip synonyms
    :return: uuid (str), name (str), syns (set, includes name, excludes uuid)
    """
    ns = find_ns(flow.nsmap, 'Flow')
    uid = str(find_common(flow, 'UUID')).strip()
    syns = set()
    name = grab_flow_name(flow, ns=ns)
    syns.add(name)
    cas = str(find_tag(flow, 'CASNumber', ns=ns)).strip()
    if cas != '':
        syns.add(cas)
    if skip_syns:
        print('Skipping synonyms')
    elif uid in ilcd_bad_synonyms:
        print('Skipping Synonyms for %s' % name)
    else:
        for syn in find_tags(flow, 'synonyms', ns='common'):
            for x in str(syn).split(';'):
                if x.strip() != '' and x.strip().lower() != 'wood':
                    syns.add(x.strip())
    return uid, name, syns



"""
This file tests BasicArchive operations that do not involve creating or adding entities-- only basic construction
of the archive and definition of catalog names.

For testing of actual entity-containing archives, look to test_base
"""

import unittest
import os
from uuid import uuid4

from antelope import local_ref, CatalogRef
from ..entity_store import SourceAlreadyKnown
from ..basic_archive import BasicArchive

WORKING_FILE = os.path.join(os.path.dirname(__file__), 'test-basic-archive.json')
conflict_file = '/dummy/conflict/file'
test_ref = 'test.basic'
test_conflict = 'test.conflict'

test_elaborated_ref = 'test.basic.my.sub.version'

archive_json = {
  "@context": "https://bkuczenski.github.io/lca-tools-datafiles/context.jsonld",
  "catalogNames": {
    test_ref: [
      WORKING_FILE
    ]
  },
  "dataReference": test_ref,
  "dataSource": WORKING_FILE,
  "dataSourceType": "BasicArchive",
  "flows": [],
  "quantities": []
}


def setUpModule():
    ar = BasicArchive(WORKING_FILE, ref=test_ref)
    ar.write_to_file(WORKING_FILE, gzip=False)


def tearDownModule():
    os.remove(WORKING_FILE)


class BasicArchiveTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ar = BasicArchive(WORKING_FILE)

    def test_rewrite_name(self):
        """
        Discovered-- if we create an archive from an existing file, but without specifying the ref, then the
        EntityStore will convert that file to a local ref and use it.  But if the file contains a ref specification,
        we want that one to win. So we use it instead.
        :return:
        """
        self.assertEqual(self.ar.ref, local_ref(WORKING_FILE))
        self.ar.load_from_dict(archive_json)
        self.assertEqual(self.ar.ref, test_ref)

    def test_catalog_ref(self):
        my_id = str(uuid4())
        self.ar.add(CatalogRef('bogus.origin', my_id, entity_type='flow'))
        self.assertEqual(self.ar[my_id].uuid, my_id)

    def test_conflicting_ref(self):
        """
        It's an error to instantiate an existing source with a new reference (why? because the source should know its
        own reference?).  If it is desired to load a source without knowing its reference, use BasicArchive.from_file()
        :return:
        """
        a = BasicArchive(WORKING_FILE, ref=test_conflict)
        with self.assertRaises(SourceAlreadyKnown):
            a.load_from_dict(archive_json)

    def test_generic_ref(self):
        """
        If a static resource is created with a fully-qualified ref, (base.ref.index.YYYYMMDD) and then it is later
        instantiated with a less-specific ref (base.ref), that escalation should be permitted (i.e. when the source
        file with the embedded ref is loaded, the more specific ref should be ignored)
        :return:
        """
        a = BasicArchive(WORKING_FILE, ref=test_ref)
        with self.assertRaises(SourceAlreadyKnown):
            a.add_new_source(test_conflict, WORKING_FILE)
        a.add_new_source(test_elaborated_ref, WORKING_FILE)

    def test_conflicting_src(self):
        """
        On the other hand, one ref is allowed to have multiple sources so this should not cause any issues
        why are we allowed to mix sources in the same ref? because a ref is a data synthesis.
        :return:
        """
        a = BasicArchive(conflict_file, ref=test_ref)
        a.load_from_dict(archive_json)
        self.assertSetEqual(set(k for k in a.get_sources(test_ref)), {conflict_file, WORKING_FILE})


if __name__ == '__main__':
    unittest.main()



from collections import defaultdict

from antelope import IndexInterface, comp_dir, CONTEXT_STATUS_
from .basic import BasicImplementation
from ..contexts import NullContext


class IndexImplementation(BasicImplementation, IndexInterface):
    """
    A CatalogInterface provides basic-level semantic data about entities

    Only requires the abstract ArchiveImplementation

    Attribute requirements for the archive:
     - everything required by BasicImplementation
     - entities_by_type()
     - count_by_type()
     - search(), including implicitly _narrow_search()
    """
    def __init__(self, *args, cutoffs=False, **kwargs):
        super(IndexImplementation, self).__init__(*args, **kwargs)
        if not hasattr(self._archive, 'ti'):
            self.re_index(cutoffs=cutoffs)

    @property
    def _terminations(self):
        return self._archive.ti

    def re_index(self, cutoffs=False):
        self._archive.ti = defaultdict(set)
        self._index_terminations(cutoffs=cutoffs)

    def _index_terminations(self, cutoffs=False):
        """
        This can't be done on add because new processes may get stored before their references are setup.
        This should only be run if the archive is local
        :param cutoffs: [False] if true, terminations include all cutoff [null-termination] exchanges, not just
          references
         :NOTE: until context refactor there is no consistent usage of terminations, so we will skip this
        :return:
        """
        for p in self._archive.entities_by_type('process'):
            if cutoffs and CONTEXT_STATUS_ == 'new':
                for x in p.exchanges():
                    if x.termination is None:
                        self._terminations[x.flow.external_ref].add((x.direction, p))
            else:
                for rx in p.reference_entity:
                    self._terminations[rx.flow.external_ref].add((rx.direction, p))
    """
    CatalogInterface core methods
    These are the main tools for describing information about the contents of the archive
    """
    def count(self, entity_type, **kwargs):
        return self._archive.count_by_type(entity_type)

    def processes(self, literal=False, **kwargs):
        for p in self._archive.search('process', **kwargs):
            yield p

    def flows(self, **kwargs):
        for f in self._archive.search('flow', **kwargs):
            yield f

    def quantities(self, **kwargs):
        for q in self._archive.search('quantity', **kwargs):
            yield q

    def lcia_methods(self, **kwargs):
        for q in self._archive.search('quantity', **kwargs):
            if q.is_lcia_method:
                yield q

    def targets(self, flow_ref, direction=None, **kwargs):
        """
        Generate processes in the archive that terminate a given exchange i.e. - have the same flow and a complementary
        direction.  If refs_only is specified, only report processes that terminate the exchange with a reference
        exchange.
        :param flow_ref: flow, exchange, or flow's external key
        :param direction: [None] filter
        :return:
        """
        if hasattr(flow_ref, 'entity_type'):
            if flow_ref.entity_type == 'flow':
                flow_ref = flow_ref.external_ref
            elif flow_ref.entity_type == 'exchange':
                if direction is None:
                    direction = flow_ref.direction
                flow_ref = flow_ref.flow.external_ref

        cdir = comp_dir(direction)
        if not isinstance(flow_ref, str) and not isinstance(flow_ref, int):  # NSUUID archives can use integer ids
            flow_ref = flow_ref.external_ref
        for x in self._terminations[flow_ref]:  # defaultdict, so no KeyError
            if direction is None:
                yield x[1]
            else:
                if cdir == x[0]:
                    yield x[1]

    def flowables(self, **kwargs):
        return self._archive.tm.flowables(**kwargs)

    def contexts(self, **kwargs):
        return self._archive.tm.contexts(**kwargs)

    def get_context(self, term, **kwargs):
        """
        I think this needs to be moved into the quantity interface
        :param term:
        :param kwargs:
        :return:
        """
        cx = self._archive.tm[term]
        if cx is None:
            return NullContext
        return cx

    '''
    def mix(self, flow_ref, direction):
        if not isinstance(flow_ref, str):
            flow_ref = flow_ref.external_ref
        terms = [t for t in self.terminate(flow_ref, direction=direction)]
        flow = self[flow_ref]
        p = LcProcess.new('Market for %s' % flow['Name'], Comment='Auto-generated')
        p.add_exchange(flow, comp_dir(direction), value=float(len(terms)))
        p.add_reference(flow, comp_dir(direction))
        for t in terms:
            p.add_exchange(flow, direction, value=1.0, termination=t.external_ref)
        return p
    '''

    '''
    def terminate(self, flow, direction=None, **kwargs):
        for p in self._archive.terminate(flow, direction=direction, **kwargs):
            yield self.make_ref(p)

    def originate(self, flow, direction=None, **kwargs):
        for p in self._archive.originate(flow, direction=direction, **kwargs):
            yield self.make_ref(p)

    def mix(self, flow, direction, **kwargs):
        return self._archive.mix(flow, direction, **kwargs)
    '''
    def unmatched_flows(self, flows, **kwargs):
        """
        Takes in a list of flows and generates a sublist of flows that were not recognized as synonyms to any local
        flows.

        Uses TermManager.get_flowable which tries the best of flow.synonyms, flow.name, str(flow)

        :param flows:
        :param kwargs:
        :return:
        """
        for flow in flows:
            try:
                self._archive.tm.get_flowable(flow)
            except KeyError:
                yield flow



from antelope import EntityNotFound, NoAccessToEntity


class BasicImplementation(object):
    def __init__(self, archive, **kwargs):
        """
        Provides common features for an interface implementation: namely, an archive and a privacy setting. Also
        provides access to certain common methods of the archive.  This should be the base class for interface-specific
        implementations.

        Requires an archive with the following attributes:
         - ref - report semantic reference
         - source - report physical data source
         - static - a boolean indicating whether the full contents of the archive are loaded into memory
         - get_uuid() - deprecated - only present for compatibility reasons
         - __getitem__ - retrieve already-loaded entity
         - retrieve_or_fetch_entity() - _fetch abstract method must be implemented

        All of these requirements are met by the standard ArchiveImplementation, with the exception of the _fetch
        abstract method.

        :param archive: an LcArchive
        :param privacy: No longer used. Privacy is enforced at the server and not the resource (where it was phony
        from the beginning)
        """
        self._archive = archive

    def validate(self):
        """
        way to check that a query implementation is valid without querying anything
        :return:
        """
        return self.origin

    @property
    def origin(self):
        return self._archive.ref

    def __str__(self):
        return '%s for %s (%s)' % (self.__class__.__name__, self.origin, self._archive.source)

    def __getitem__(self, item):
        return self._archive[item]

    def _dereference_entity(self, external_ref):
        if hasattr(external_ref, 'external_ref'):
            eref = external_ref
            external_ref = eref.external_ref
        else:
            eref = None
        entity = self.get(external_ref)
        # if entity:
        if not entity.is_entity:
            raise NoAccessToEntity(entity.link)
        return entity

    def get_item(self, external_ref, item):
        """
        In this, we accept either an external_ref or an entity reference itself.  If the latter, we dereference via
        the archive to an actual entity, which we then ask for the item.  If the dereference and the reference are the
        same, throws an error.
        :param external_ref:
        :param item:
        :return:
        """
        entity = self._dereference_entity(external_ref)
        if entity.has_property(item):
            return entity[item]
        raise KeyError('%s: %s [%s]' % (self.origin, external_ref, item))
        # raise EntityNotFound(external_ref)

    def get_reference(self, key):
        entity = self._dereference_entity(key)
        if entity is None:
            return None
        if entity.entity_type == 'process':
            # need to get actual references with exchange values-- not the reference_entity
            return [x for x in entity.references()]
        return entity.reference_entity

    def get_uuid(self, external_ref):
        u = self._archive.get_uuid(external_ref)
        if u is None:
            return False
        return u

    def _fetch(self, external_ref, **kwargs):
        if external_ref is None:
            return None
        if self._archive.static:
            return self._archive[external_ref]
        try:
            return self._archive.retrieve_or_fetch_entity(external_ref, **kwargs)
        except NotImplementedError:
            return None

    '''
    def lookup(self, external_ref, **kwargs):
        if self._fetch(external_ref, **kwargs) is not None:
            return True
        return False
    '''

    def get(self, external_ref, **kwargs):
        """
        :param external_ref: may also be link, as long as requested origin is equal or lesser in specificity
        :param kwargs:
        :return: entity or None
        """
        e = self._fetch(external_ref, **kwargs)
        if e is not None:
            return e
        if isinstance(external_ref, int):  # we got away with this before by falling back on NSUUIDs
            external_ref = str(external_ref)
        e = self._fetch(external_ref, **kwargs)
        if e is not None:
            return e
        er_s = external_ref.split('/')
        if self.origin.startswith(er_s[0]):
            e = self._fetch('/'.join(er_s[1:]), **kwargs)
            if e is not None:
                return e
        raise EntityNotFound(external_ref)

    def synonyms(self, item, **kwargs):
        return self._archive.tm.synonyms(item)

    def properties(self, external_ref, **kwargs):
        e = self._dereference_entity(external_ref)
        for i in e.properties():
            yield i



import os
import re

from antelope import EntityNotFound, local_ref

from ..lc_resource import LcResource
from ..lcia_engine import LciaDb

from ..archives import InterfaceError


class Configurator(object):

    _cat_root = None

    def _add_config(self, config, *args):
        """
        does no validation
        :param config:
        :param args:
        :return:
        """
        self._config[config].add(args)

    '''Catalog Emulation'''
    def abs_path(self, rel_path):
        if os.path.isabs(rel_path):
            return rel_path
        elif rel_path.startswith('$CAT_ROOT'):
            return re.sub('^\$CAT_ROOT', self.catalog_root, rel_path)
        return os.path.join(self.catalog_root, rel_path)

    @staticmethod
    def cache_file(source):
        return os.path.join('/this/path/does/not/exist', os.path.basename(source))

    @property
    def catalog_root(self):
        return self._cat_root

    @catalog_root.setter
    def catalog_root(self, root):
        if os.path.isdir(root):
            self._cat_root = root
        else:
            raise ValueError('Catalog root does not exist or is not directory: %s' % root)

    '''
    Class Methods for creating new configurations
    '''

    @classmethod
    def new(cls, source, ds_type, **kwargs):
        ref = local_ref(source, prefix='config')
        res = LcResource(ref, source, ds_type, **kwargs)
        return cls(res)

    @classmethod
    def from_download(cls, url, md5sum=None):
        # punt on this because of stupid mkstemp stuff
        pass

    @classmethod
    def from_existing_resource(cls, catalog_root, ref, interface='basic'):
        resource_file = os.path.join(catalog_root, 'resources', ref)
        ress = [k for k in LcResource.from_json(resource_file) if interface in k.interfaces]
        if len(ress) > 1:
            print('Warning: using first of several matching resources')
        cfg = cls(ress[0], catalog_root=catalog_root)
        return cfg

    def __init__(self, resource, catalog_root=None):
        if catalog_root is not None:
            self.catalog_root = catalog_root
        self._ldb = LciaDb.new()
        self._resource = resource
        self._resource.check(self)
        self._config = resource.config

    @property
    def archive(self):
        return self._resource.archive

    @property
    def lcia_engine(self):
        return self._ldb.tm

    def load_all(self):
        self._resource.archive.load_all()

    def add_hint(self, hint_type, local_name, canonical_name):
        return self.add_config('hints', hint_type, local_name, canonical_name)

    def add_config(self, option, *args):
        """
        Add a configuration setting to the resource, and apply it to the archive.
        :param option: the option being configured
        :param args: the arguments, in the proper sequence
        :return: None if archive doesn't support configuration; False if unsuccessful, True if successful
        """
        try:
            cf = self.archive.make_interface('configure')
        except InterfaceError:
            print('No Configure interface')
            return None

        if cf.check_config(option, args):
            if option == 'hints':
                self.lcia_engine.apply_hints(self.archive.catalog_names, [args])
            cf.apply_config({option: {args}})
            self._add_config(option, *args)
            return True
        print('Configuration failed validation.')
        return False

    def _serialize_config(self):
        j = dict()
        for k, v in self._config.items():
            j[k] = sorted([list(g) for g in v], key=lambda x: x[0])
        return j

    def write_to_catalog(self, assign_ref=None, catalog_root=None):
        if catalog_root is None:
            if self.catalog_root is None:
                raise ValueError('Catalog root must be specified')
            catalog_root = self.catalog_root
        resource_root = os.path.join(catalog_root, 'resources')
        if assign_ref is None:
            assign_ref = self._resource.reference
        self._resource.write_to_file(resource_root, assign_ref, apply_config=self._config)

    def check_contexts(self):
        """
        Compares contexts known to the resource with contexts known to the captive LCIA engine.  Prints matching
        contexts. Returns a list of non-matching contexts.
        :return:
        """
        valid = []
        null = []
        for cx in self._resource.make_interface('index').contexts():
            if self._ldb.tm[cx] is None:
                null.append(cx)
            else:
                valid.append(cx)
        print('Valid Contexts:')
        for k in valid:
            print(' %s ==> %s' % (k.name, repr(self._ldb.tm[k])))

        '''
        print('Null Contexts:')
        for k in null:
            print(' %s' % k.fullname)
        '''
        return null

    def check_quantities(self):
        """
        Compares quantities known to the resource with quantities known to the captive LCIA engine.  Prints recognized
        quantities. Returns a list of unknown quantities.
        :return:
        """
        valid = []
        null = []
        for q in self.archive.entities_by_type('quantity'):
            try:
                can = self.lcia_engine.get_canonical(q)
                valid.append((q, can))
            except EntityNotFound:
                null.append(q)
        print ('Known Quantities:')
        for k in valid:
            print(' %s ==> %s' % (k[0].name, k[1].name))

        '''
        print('Unrecognized Quantities:')
        for k in null:
            print(' %s' % k)
        '''
        return null

    def unmatched_flowables(self):
        """
        Compares flowables known to the resource with flowables known to the captive LCIA engine.  Returns a list of
        non-matching flowables (does not print known flowables).
        :return:
        """
        null = self._ldb.tm.unmatched_flowables(self._resource.make_interface('index').flowables())
        print('%d Unmatched flowables' % len(null))
        return null



import unittest
import os

from .. import LcCatalog

from ...archives import LcArchive

cat = LcCatalog.make_tester()

blast_furnace = '00420798-e9d1-4de9-8745-09bd85f31db8'
blast_furnace_gas = 'b254bbdf-fb2b-4878-aec9-2a7820f3f32e'

alt_archive_path = os.path.join(os.path.dirname(__file__), 'blast_furnace_test.json.gz')
ar = LcArchive.from_file(alt_archive_path)
test_ref = ar.ref

cat.add_existing_archive(ar, interfaces='inventory', store=False)


class ProcessRefTest(unittest.TestCase):
    """
    Things to test:
     * default_rx is properly set for processes with one reference exchange
     * default_rx is not set for processes with zero or more than one ref ex
     * RxRef acts like a reference exchange:
       - is_reference is True
       - membership in both process_ref.reference_entity and process.reference_entity
     * is_allocated behaves as expected

    """
    @classmethod
    def setUpClass(cls):
        cls._ar = ar

    def test_rxref_is_ref(self):
        """
        The RxRefs that populate a process_ref.reference_entity should test as being contained in the originating
        process.reference_entity.
        :return:
        """
        p = self._ar.retrieve_or_fetch_entity(blast_furnace)
        p_ref = p.make_ref(cat.query(test_ref))
        for rx in p_ref.reference_entity:
            self.assertTrue(p.has_reference(rx))

    @staticmethod
    def _get_matching_ref(rx, p):
        for rxc in p.references():
            if rxc.key == rx.key:
                return rxc
        raise AssertionError('no matching ref found')

    def test_p_ref_access_ref_by_string(self):
        p_ref = cat.query(test_ref).get(blast_furnace)
        rx1 = p_ref.reference(blast_furnace_gas)
        rx2 = p_ref.reference(cat.query(test_ref).get(blast_furnace_gas))
        self.assertIs(rx1, rx2)

    '''
    def test_p_ref_is_allocated(self):
        """
        is_allocated should go away
        :return:
        """
        p = self._ar.retrieve_or_fetch_entity(blast_furnace)
        p_ref = cat.query(test_ref).get(blast_furnace)
        for rxref in p_ref.reference_entity:
            rx = self._get_matching_ref(rxref, p)
            self.assertEqual(p_ref.is_allocated(rx), p_ref.is_allocated(rxref))
    '''


if __name__ == '__main__':
    unittest.main()



"""
The CLookup is a classic example of trying to solve a very complex data problem with a very complex data structure.
It does not entirely succeed, but it works well enough as a stopgap until I have a chance to give an honest crack at
a graph database.

Flow characterization is the very complex data problem: the "quantity relation" maps a large, diverse set of inputs:
 - flowable (substance)
 - reference quantity (in our data system, tied to flowable)
 - query quantity
 - context / compartment
 - location / locale
to a numeric output, namely the amount of the query quantity that corresponds to a unit of the reference quantity.
Technically, this amount also has uncertainty / other quantitative characteristics.

The idea behind a CLookup is that it contains all known characterizations for a given flowable [substance] with respect
to a given quantity.  The CLookup is selected by specifying the flowable and the quantity, and then the CLookup is used
to retrieve a set of Characterization objects for a given context (hence the 'C' in 'CLookup'). The characterization
already stores a mapping of locale to factor, and also stores the flowable (with its native reference quantity used
to interpret the factor).

So it solves a very narrow portion of the problem and leaves a lot to outside code.

The dream would be to design a graph database that held all of these parameters and magically obtained all the factors
that applied to a given query-- that graph database would replace the current Term Manager and everything else under
its hood. But first we will learn to walk...
"""
from ..contexts import Context, NullContext


class QuantityMismatch(Exception):
    pass


class DuplicateOrigin(Exception):
    pass


class CLookup(object):
    """
    A CLookup is a kind of fuzzy dictionary that maps context to best-available characterization. A given CLookup is
    associated with a single quantity and a specific flowable.  The query then provides the compartment and returns
    either: a set of best-available characterizations; a single characterization according to a selection rule; or a
    characterization factor (float) depending on which method is used.
    """
    def __init__(self):
        self._dict = dict()
        self._q = None

    def __repr__(self):
        return '%s(%s: %d contexts)' % (self.__class__.__name__, self._q, len(self._dict))

    def __getitem__(self, item):
        """
        Returns
        :param item:
        :return:
        """
        if item is None:
            item = NullContext
        if not isinstance(item, Context):
            raise TypeError('Supplied CLookup key is not a Context: %s (%s)' % (item, type(item)))
        if item in self._dict:
            return self._dict[item]
        return set()

    def _check_qty(self, cf):
        if self._q is None:
            self._q = cf.quantity
        else:
            if cf.quantity != self._q:
                raise QuantityMismatch('Inbound: %s\nCurrent: %s' % (cf.quantity, self._q))

    def add(self, value, key=None):
        if key is None:
            key = value.context
        if isinstance(key, Context):
            self._check_qty(value)
            if key not in self._dict:
                self._dict[key] = set()
            '''
            if any(k.origin == value.origin for k in self._dict[key]):
                raise DuplicateOrigin(value.origin)
            '''
            self._dict[key].add(value)
        else:
            raise ValueError('Context is not valid: %s (%s)' % (key, type(key)))

    def remove(self, value):
        key = value.context
        self._dict[key].remove(value)

    def keys(self):
        return self._dict.keys()

    def cfs(self):
        for c, cfs in self._dict.items():
            for cf in cfs:
                yield cf

    def _context_origin(self, item, origin):
        if origin is None:
            return list(self.__getitem__(item))
        else:
            return [k for k in self.__getitem__(item) if k.origin == origin]

    def find(self, item, dist=1, return_first=True, origin=None):
        """
        Hunt for a matching compartment. 'dist' param controls the depth of search:
          dist = 0: equivalent to __getitem__
          dist = 1: also check compartment's children (subcompartments), to any depth, returning all CFs encountered
            (unless return_first is True, in which case all CFs from the first nonempty compartment found are returned)
          dist = 2: also check compartment's parent
          dist = 3: also check all compartment's parents until root. Useful for finding unit conversions.
        By default (dist==1), checks compartment self and children. Returns a set.
        :param item: a Compartment
        :param dist: how far to search (with limits) (default: 1= compartment + children)
        :param return_first: stop hunting as soon as a cf is found
        :param origin: [None] if present, only return cfs whose origins match the specification
        :return: a list of characterization factors that meet the query criteria, ordered by increasing dist
        """
        #!TODO: This should be performed by the term manager, not the CLookup (though that is the whole point of CLookup)
        if not isinstance(item, Context):
            return []

        def found(res):
            return len(res) > 0 and return_first
        results = self._context_origin(item, origin)
        if found(results):
            return results

        if dist > 0:
            for s in item.self_and_subcompartments:  # note: this is depth first
                if s is item:
                    continue  # skip self, just recurse subcompartments
                results += self._context_origin(s, origin)
                if found(results):
                    return results

        if dist > 1:
            item = item.parent
            results += self._context_origin(item, origin)
            if found(results):
                return results

        while dist > 2 and item is not None:
            item = item.parent
            results += self._context_origin(item, origin)
            if found(results):
                return results

        return results

    def find_first(self, item, dist=3):
        cfs = self.find(item, dist=dist, return_first=True)
        return list(cfs)[0]

    @staticmethod
    def _ser_set(cf_set, values=False):
        return {cf.origin: cf.serialize(values=values, concise=True) for cf in cf_set}

    def serialize(self, values=False):
        return {str(c): self._ser_set(cfs, values=values) for c, cfs in self._dict.items()}

    def serialize_for_origin(self, origin, values=False):
        """
        Note: In the event that the CLookup includes multiple CFs for the same flowable and the same origin, only the
        first (at random) will be included, because originally I had disallowed multiple CFs for the same origin.
        :param origin:
        :param values:
        :return:
        """
        cxs = dict()
        for c, cfs in self._dict.items():
            try:
                cf_filt = next(cf for cf in cfs if cf.origin == origin)  # duplicate entries per origin not allowed
            except StopIteration:
                continue
            cxs[str(c)] = cf_filt.serialize(values=values, concise=True)
        return cxs


class FactorCollision(Exception):
    pass


class SCLookup(CLookup):
    """
    A Strict CLookup that permits only one CF to be stored per compartment and raises an error if an additional one
    is added.
    """
    def add(self, value, key=None):
        if key is None:
            key = value.context
        if key in self._dict and len(self._dict[key]) > 0:
            existing = list(self._dict[key])[0]
            if existing.value == value.value:
                return
            print('Collision with context: %s' % repr(key))
            print(repr(value))
            print('%s current' % repr(existing))
            raise FactorCollision('This context already has a CF defined!')
        super(SCLookup, self).add(value, key)



"""
This file deals with tests for auto-loading and masquerading of LCIA methods
"""
import unittest

from .. import LcCatalog
from ...data_sources.local import make_config


cat = LcCatalog.make_tester()
cfg = make_config('ipcc2007')
ref = next(cfg.references)


def setUpModule():
    if ref not in cat.references:
        cat.add_resource(next(cfg.make_resources(ref)))


class QuantityRefTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.gwp = next(cat.query(ref).lcia_methods(Name='Global Warming'))  # canonical
        cls.gwp_ref = cat._qdb[cls.gwp.external_ref]  # original ref
        cls.gwp_true = cat.get_archive(cls.gwp_ref.origin).get(cls.gwp_ref.external_ref)  # authentic entity

    def test_origins(self):
        self.assertEqual(self.gwp.origin, ref)
        self.assertEqual(self.gwp_ref.origin, ref)
        res = cat.get_resource(ref)
        self.assertEqual(self.gwp_true.origin, res.archive.names[res.source])

    def test_factors(self):
        self.assertEqual(len([k for k in self.gwp.factors()]), 91)
        self.assertEqual(len([k for k in self.gwp_ref.factors()]), 91)
        self.assertEqual(len([k for k in self.gwp_true.factors()]), 91)

    def test_properties(self):
        self.assertTrue(self.gwp.has_property('indicator'))
        self.assertTrue(self.gwp_ref.has_property('indicator'))
        self.assertTrue(self.gwp_true.has_property('indicator'))




from ..xml_widgets import render_text_block, find_tag
from lxml import objectify
import unittest


ecospold2_sample = '''
<ecoSpold xmlns="http://www.EcoInvent.org/EcoSpold02">
  <childActivityDataset>
    <activityDescription>
      <activity id="aed7f674-fc6b-5c7e-9e09-0b89adb92259" activityNameId="202fa2d7-8c4e-4e6a-8870-d8db89b890fa" activityNameContextId="de659012-50c4-4e96-b54a-fc781bf987ab" parentActivityId="aed7f674-fc6b-5c7e-9e09-0b89adb92259" inheritanceDepth="-1" type="1" specialActivityType="1">
        <activityName xml:lang="en">market for electricity, medium voltage</activityName>
        <includedActivitiesStart xml:lang="en">This activity starts from 1kWh of electricity fed into the medium voltage transmission network.</includedActivitiesStart>
        <includedActivitiesEnd xml:lang="en">This activity ends with the transport of 1 kWh of medium voltage electricity in the transmission network over aerial lines and cables.

This dataset includes: 
  - electricity inputs produced in this country and from imports and transformed to medium voltage
 - the transmission network
 - direct emissions to air (SF6) 
 - electricity losses during transmission 

This dataset doesn't include
 - electricity losses during transformation from high to medium voltage or medium to low, as these are included in the dataset for transformation
  - leakage of insulation oil from cables and electro technical equipment (transformers, switchgear, circuit breakers) because this only happens in case of accidental release
 - SF6 emissions during production and deconstruction of the switchgear, as these are accounted for in the transmission network dataset.
- Emissions to soil from leakage of treating substances from poles (included in transmission network).</includedActivitiesEnd>
        <generalComment>
          <text xml:lang="en" index="10000">The shares of electricity technologies on this market are valid for the year 2014. They have been implemented by the software layer and don't represent the production volumes in the unlinked datasets valid for the year 2012. These shares have been calculated based on statistics from 2014. Basic source: Statistical tables by StatCan on electricity production in Canada. Further information will be available in a report available on ecoQuery. </text>
          <variable xml:lang="en" name="year">2014</variable>
          <text xml:lang="en" index="1">This dataset describes the electricity available on the medium voltage level in {{location}} for year {{year}}. This is done by showing the transmission of 1kWh electricity at medium voltage.</text>
          <variable xml:lang="en" name="location">Canada, Québec</variable>
        </generalComment>
      </activity>
    </activityDescription>
  </childActivityDataset>
</ecoSpold>
'''


converted_string_ecospold = '''This dataset describes the electricity available on the medium voltage level in Canada, Québec for year 2014. This is done by showing the transmission of 1kWh electricity at medium voltage.
The shares of electricity technologies on this market are valid for the year 2014. They have been implemented by the software layer and don't represent the production volumes in the unlinked datasets valid for the year 2012. These shares have been calculated based on statistics from 2014. Basic source: Statistical tables by StatCan on electricity production in Canada. Further information will be available in a report available on ecoQuery. '''


class XmlWidgetTestCase(unittest.TestCase):
    def test_objectify(self):
        o = objectify.fromstring(ecospold2_sample)
        self.assertEqual(len(o), 1)
        self.assertEqual(len(o['childActivityDataset']), 1)

    def test_find_tags_ecospold(self):
        o = objectify.fromstring(ecospold2_sample)
        self.assertEqual(len(find_tag(o, 'generalComment')), 1)

    def test_string_substitution_ecospold(self):
        o = objectify.fromstring(ecospold2_sample)
        self.assertEqual(render_text_block(find_tag(o, 'generalComment')), converted_string_ecospold)


if __name__ == '__main__':
    unittest.main()



import unittest
from .base_testclass import BasicEntityTest


class FlowsTest(BasicEntityTest):
    def test_synonyms(self):
        f = self.A["018872d0-ebce-3dfe-9552-3ff163fb468c"]
        self.assertIn(f.name, self.A.tm.synonyms(f.link))


if __name__ == '__main__':
    unittest.main()



"""
The LcCatalog provides a semantic interface to a collection of (local and remote) read-only LcArchives, which provide
access to physical data.

It is made up of the following components:

  * built on an LciaEngine
  + local, persistent storage of resources, indexes, cache data + etc
  + A resolver, which translates semantic references into resources.  Input: semantic ref. output: CatalogInterface.
  + an interface generator, which creates archive accessors on demand based on resource information from the resolver
  x An internal cache of entities retrieved, by full reference-- this has been cut

From the catalog_ref file, the catalog should meet the following spec:
          Automatic - entity information
           catalog.query(origin) - returns a query interface
           catalog.lookup(origin, external_ref) - returns the origin of the lowest-priority resource resolving the ref
           catalog.fetch(origin, external_ref) - return a reference to the object that can be queried + handled

          LC Queries:
           see lcatools.interfaces.*

"""

import os
# import re
import hashlib
# from collections import defaultdict

from ..archives import InterfaceError
from ..lcia_engine import LciaDb


from antelope import CatalogRef, UnknownOrigin  # , EntityNotFound
from ..catalog_query import CatalogQuery, INTERFACE_TYPES, zap_inventory
from .lc_resolver import LcCatalogResolver
from ..lc_resource import LcResource
# from lcatools.flowdb.compartments import REFERENCE_INT  # reference intermediate flows


class DuplicateEntries(Exception):
    pass


class CatalogError(Exception):
    pass


class StaticCatalog(object):

    """
    Provides query-based access to LCI information. The static version is ideal for creating read-only web resources
    from curated LcCatalogs. However, it must already exist. Only an LcCatalog (or subclasses) support de novo
    instantiation.

    A catalog is stored in the local file system and creates and stores resources relative to its root directory.
    Subfolders (all accessors return absolute paths):
    Public subfolders:
     LcCatalog.resource_dir
     LcCatalog.archive_dir

    Public filenames:
     LcCatalog.cache_file(src) returns a sha1 hash of the source filename in the [absolute] cache dir
     LcCatalog.download_file(src) returns a sha1 hash of the source filename in the [absolute] download dir

    Private folders + files:
     LcCatalog._download_dir
     LcCatalog._index_dir
     LcCatalog._index_file(src) returns a sha1 hash of the source filename in the [absolute] index dir
     LcCatalog._cache_dir
     LcCatalog._entity_cache: local entities file in root
     LcCatalog._reference_qtys: reference quantities file in root
     LcCatalog._compartments: local compartments file (outmoded in Context Refactor)


    """
    @property
    def resource_dir(self):
        return os.path.join(self._rootdir, 'resources')

    @property
    def _download_dir(self):
        return os.path.join(self._rootdir, 'downloads')

    @staticmethod
    def _source_hash_file(source):
        """
        Creates a stable filename from a source argument.  The source is the key found in the _archive dict, and
        corresponds to a single physical data source.  The filename is a sha1 hex-digest, .json.gz
        :param source:
        :return:
        """
        h = hashlib.sha1()
        h.update(source.encode('utf-8'))
        return h.hexdigest()

    @property
    def _index_dir(self):
        return os.path.join(self._rootdir, 'index')

    def _index_file(self, source):
        return os.path.join(self._index_dir, self._source_hash_file(source) + '.json.gz')

    @property
    def _cache_dir(self):
        return os.path.join(self._rootdir, 'cache')

    def cache_file(self, source):
        return os.path.join(self._cache_dir, self._source_hash_file(source) + '.json.gz')

    @property
    def archive_dir(self):
        return os.path.join(self._rootdir, 'archives')

    '''
    @property
    def _entity_cache(self):
        return os.path.join(self._rootdir, 'entity_cache.json')
    '''

    @property
    def _reference_qtys(self):
        return os.path.join(self._rootdir, 'reference-quantities.json')

    '''
    @property
    def _compartments(self):
        """
        Deprecated
        :return:
        """
        return os.path.join(self._rootdir, 'local-compartments.json')
    '''

    @property
    def _contexts(self):
        return os.path.join(self._rootdir, 'local-contexts.json')

    @property
    def _flowables(self):
        return os.path.join(self._rootdir, 'local-flowables.json')

    def _localize_source(self, source):
        if source is None:
            return None
        if source.startswith(self._rootdir):
            #return re.sub('^%s' % self._rootdir, '$CAT_ROOT', source)
            # Should work on both mac and windows
            return os.path.join('$CAT_ROOT', os.path.relpath(source, self._rootdir))
        return source

    def abs_path(self, rel_path):
        if os.path.isabs(rel_path):
            return rel_path
        elif rel_path.startswith('$CAT_ROOT'):
            #return re.sub('^\$CAT_ROOT', self.root, rel_path)
            # Should work on both mac and windows
            return os.path.abspath(os.path.join(self.root, os.path.relpath(rel_path, '$CAT_ROOT')))
        return os.path.abspath(os.path.join(self.root, rel_path))

    @property
    def root(self):
        return self._rootdir

    def __init__(self, rootdir, strict_clookup=True, **kwargs):
        """
        Instantiates a catalog based on the resources provided in resource_dir
        :param rootdir: directory storing LcResource files.
        :param strict_clookup: [True] whether to enforce uniqueness on characterization factors (raise an error when a
         non-matching duplicate characterization is encountered). If False, selection among conflicting factors is
         not well defined and may be done interactively or unpredictably
        :param kwargs: passed to Qdb
        """
        self._rootdir = os.path.abspath(rootdir)
        if not os.path.exists(self._rootdir):
            raise FileNotFoundError(self._rootdir)
        self._resolver = LcCatalogResolver(self.resource_dir)

        """
        _archives := source -> archive
        _names :=  ref:interface -> source
        _nicknames := nickname -> source
        """
        self._nicknames = dict()  # keep a collection of shorthands for sources

        self._queries = dict()  # keep a collection of CatalogQuery instances for each origin

        '''
        LCIA: 
        '''
        qdb = LciaDb.new(source=self._reference_qtys, contexts=self._contexts, flowables=self._flowables,
                         strict_clookup=strict_clookup, **kwargs)
        self._qdb = qdb
        res = LcResource.from_archive(qdb, interfaces=('index', 'quantity'), store=False)
        self._resolver.add_resource(res, store=False)

    '''
    The thing that distinguishes a catalog from an archive is its centralized handling of quantities via the qdb
    '''
    @property
    def qdb(self):
        """
        Provides query access to the quantity database. Should be like cat.query('local.qdb'), except that
        it provides a basic query- which is what internal quantities use themselves
        :return:
        """
        return self._qdb.query

    @property
    def lcia_engine(self):
        return self._qdb.tm

    def register_quantity_ref(self, q_ref):
        print('registering %s' % q_ref.link)
        self._qdb.add(q_ref)

    @property
    def sources(self):
        for k in self._resolver.sources:
            yield k

    @property
    def references(self):
        for ref, ints in self._resolver.references:
            yield ref

    @property
    def interfaces(self):
        for ref, ints in self._resolver.references:
            for i in ints:
                yield ':'.join([ref, i])

    def show_interfaces(self):
        for ref, ints in sorted(self._resolver.references):
            print('%s [%s]' % (ref, ', '.join(ints)))

    '''
    Nicknames
    '''
    @property
    def names(self):
        """
        List known references.
        :return:
        """
        for k, ifaces in self._resolver.references:
            for iface in ifaces:
                yield ':'.join([k, iface])
        for k in self._nicknames.keys():
            yield k

    def add_nickname(self, source, nickname):
        """
        quickly refer to a specific data source already present in the archive
        :param source:
        :param nickname:
        :return:
        """
        if self._resolver.known_source(source):
            self._nicknames[nickname] = source
        else:
            raise KeyError('Source %s not found' % source)


    def has_resource(self, res):
        return self._resolver.has_resource(res)

    '''
    Retrieve resources
    '''
    def _find_single_source(self, origin, interface, source=None, strict=True):
        r = self._resolver.get_resource(ref=origin, iface=interface, source=source, include_internal=False, strict=strict)
        r.check(self)
        return r.source

    def get_resource(self, name, iface=None, source=None, strict=True):
        """
        retrieve a resource by providing enough information to identify it uniquely.  If strict is True (default),
        then parameters are matched exactly and more than one match raises an exception. If strict is False, then
        origins are matched approximately and the first (lowest-priority) match is returned.

        :param name: nickname or origin
        :param iface:
        :param source:
        :param strict:
        :return:
        """
        if name in self._nicknames:
            return self._resolver.get_resource(source=self._nicknames[name], strict=strict)
        iface = zap_inventory(iface, warn=True)  # warn when requesting the wrong interface
        return self._resolver.get_resource(ref=name, iface=iface, source=source, strict=strict)

    def get_archive(self, ref, interface=None, strict=False):
        interface = zap_inventory(interface, warn=True)
        if interface in INTERFACE_TYPES:
            rc = self.get_resource(ref, iface=interface, strict=strict)
        else:
            rc = self.get_resource(ref, strict=strict)
        rc.check(self)
        return rc.archive

    '''
    Main data accessor
    '''
    def _sorted_resources(self, origin, interfaces, strict):
        for res in sorted(self._resolver.resolve(origin, interfaces, strict=strict),
                          key=lambda x: (not (x.is_loaded and x.static), x.priority, x.reference != origin)):
            yield res

    def gen_interfaces(self, origin, itype=None, strict=False):
        """
        Generator of interfaces by spec

        :param origin:
        :param itype: single interface or iterable of interfaces
        :param strict: passed to resolver
        :return:
        """
        # if itype == 'quantity':
        #    yield self._qdb.make_interface(itype)

        for res in self._sorted_resources(origin, itype, strict):
            res.check(self)
            try:
                yield res.make_interface(itype)
            except InterfaceError:
                continue

        '''
        # no need for this because qdb is (a) listed in the resolver and (b) upstream of everything
        if 'quantity' in itype:
            yield self._qdb  # fallback to our own quantity db for Quantity Interface requests
            '''

    """
    public functions -- should these operate directly on a catalog ref instead? I think so but let's see about usage
    """
    def query(self, origin, strict=False, refresh=False, **kwargs):
        """
        Returns a query using the first interface to match the origin.
        :param origin:
        :param strict: [False] whether the resolver should match the origin exactly, as opposed to returning more highly
         specified matches.  e.g. with strict=False, a request for 'local.traci' could be satisfied by 'local.traci.2.1'
         whereas if strict=True, only a resource matching 'local.traci' exactly will be returned
        :param refresh: [False] by default, the catalog stores a CatalogQuery instance for every requested origin.  With
         refresh=True, any prior instance will be replaced with a fresh one.
        :param kwargs:
        :return:
        """

        next(self._resolver.resolve(origin, strict=strict))

        if refresh or (origin not in self._queries):
            self._queries[origin] = CatalogQuery(origin, catalog=self, **kwargs)
        return self._queries[origin]

    def lookup(self, catalog_ref, keep_properties=False):
        """
        Attempts to return a valid grounded reference matching the one supplied.
        :param catalog_ref:
        :param keep_properties: [False] if True, apply incoming ref's properties to grounded ref, probably with a
        prefix or something.
        :return:
        """
        ref = self.query(catalog_ref.origin).get(catalog_ref.external_ref)
        if keep_properties:
            for k in catalog_ref.properties():
                ref[k] = catalog_ref[k]
        return ref

    '''
    def lookup(self, origin, external_ref=None):
        """
        Attempts to secure an entity
        :param origin:
        :param external_ref:
        :return: The origin of the lowest-priority resource to match the query
        """
        if external_ref is None:
            origin, external_ref = origin.split('/', maxsplit=1)
        for i in self.gen_interfaces(origin):
            if i.lookup(external_ref):
                return i.origin
        for i in self.gen_interfaces('.'.join(['foreground', origin])):
            if i.lookup(external_ref):
                return i.origin
        raise EntityNotFound('%s/%s' % (origin, external_ref))

    def fetch(self, origin, external_ref=None):
        if external_ref is None:
            origin, external_ref = origin.split('/', maxsplit=1)
        org = self.lookup(origin, external_ref)
        return self.query(org).get(external_ref)
    '''

    def catalog_ref(self, origin, external_ref, entity_type=None, **kwargs):
        """
        TODO: make foreground-generated CatalogRefs lazy-loading. This mainly requires removing the expectation of a
        locally-defined reference entity, and properly implementing and using a reference-retrieval process in the
        basic interface.
        :param origin:
        :param external_ref:
        :param entity_type:
        :return:
        """
        try:
            q = self.query(origin)
        except UnknownOrigin:
            return CatalogRef(origin, external_ref, entity_type=entity_type, **kwargs)
        return q.get(external_ref)
        # except EntityNotFound:  why would we catch this?
        #     return CatalogRef.from_query(external_ref, q, entity_type=entity_type, **kwargs)




import re

from .basic import BasicImplementation
from antelope import BackgroundInterface, ExteriorFlow, EntityNotFound, comp_dir  # , ProductFlow
from antelope_core.contexts import Context


class NonStaticBackground(Exception):
    pass


def search_skip(entity, search):
    if search is None:
        return False
    return not bool(re.search(search, str(entity), flags=re.IGNORECASE))


class BackgroundImplementation(BasicImplementation, BackgroundInterface):
    """
    The default Background Implementation exposes an ordinary inventory database as a collection of LCI results.
    Because it does not perform any ordering, there is no way to distinguish between foreground and background
    elements in a database using the proxy. It is thus inconsistent for the same resource to implement both
    inventory and [proxy] background interfaces.
    """
    def __init__(self, *args, **kwargs):
        super(BackgroundImplementation, self).__init__(*args, **kwargs)

        self._index = None

    def check_bg(self, **kwargs):
        self.setup_bm(**kwargs)
        return bool(self._index is not None)

    def setup_bm(self, index=None):
        """
        Requires an index interface or catalog query <-- preferred
        HOLD UP- catalog query limits access to interfaces named in the resource, which may not include index.
        It also returns entity references, which may be inadequate to the task of generating results (e.g. it is
        an error for the same LOCAL implementation to provide both 'exchange' and 'background' as they will be the
        same).  So routes that require exchange access, e.g. to fake up lci, should access the archive directly
        :param index:
        :return:
        """
        if self._index is None:
            if index is None:
                self._index = self._archive.make_interface('index')
            else:
                self._index = index

    def _ensure_ref_flow(self, ref_flow):
        if ref_flow is not None:
            if isinstance(ref_flow, str) or isinstance(ref_flow, int):
                ref_flow = self._archive.retrieve_or_fetch_entity(ref_flow)
        return ref_flow

    def foreground_flows(self, search=None, **kwargs):
        """
        No foreground flows in the proxy background
        :param search:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def background_flows(self, search=None, **kwargs):
        """
        all process reference flows are background flows
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for p in self._archive.entities_by_type('process'):
            for rx in p.references():
                if search_skip(p, search):
                    continue
                yield rx  # ProductFlow(self._archive.ref, rx.flow, rx.direction, p, None) don't need this

    def exterior_flows(self, direction=None, search=None, **kwargs):
        """
        Exterior flows are all flows that do not have interior terminations (i.e. not found in the index targets)
        Since contexts are still in limbo, we need a default directionality (or some way to establish directionality
        for compartments..) but for now let's just use default 'output' for all exterior flows
        :param direction:
        :param search:
        :param kwargs:
        :return:
        """
        self.check_bg()
        for f in self._index.flows():
            if search_skip(f, search):
                continue
            try:
                next(self._index.targets(f.external_ref, direction=direction))
            except StopIteration:
                cx = self._index.get_context(f.context)
                dir = comp_dir(cx.sense)
                '''
                if self.is_elementary(f):
                    yield ExteriorFlow(self._archive.ref, f, 'Output', f['Compartment'])
                else:
                    yield ExteriorFlow(self._archive.ref, f, 'Output', None)
                '''
                yield ExteriorFlow(self._archive.ref, f, dir, cx)

    def consumers(self, process, ref_flow=None, **kwargs):
        """
        Not supported for trivial backgrounds
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def dependencies(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have no dependencies
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in []:
            yield i

    def emissions(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    yield i

    def cutoffs(self, process, ref_flow=None, **kwargs):
        """
        All processes are LCI, so they have only exterior flows. Emissions are the exterior flows with non-elementary
        contexts
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        for i in self.lci(process, ref_flow=ref_flow, **kwargs):
            if isinstance(i.termination, Context):
                if i.termination.elementary:
                    continue
            yield i

    def foreground(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._index.get(process)
        yield p.reference(ref_flow)  # should be just one exchange

    def is_in_scc(self, process, ref_flow=None, **kwargs):
        """
        Distinction between is_in_background and is_in_scc will reveal the proxy nature of the interface
        :param process:
        :param ref_flow:
        :param kwargs:
        :return:
        """
        return False  # proxy has no knowledge of SCCs

    def is_in_background(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        try:
            self._archive.retrieve_or_fetch_entity(process)
        except EntityNotFound:
            return False
        return True

    def ad(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def bf(self, process, ref_flow=None, **kwargs):
        for i in []:
            yield i

    def lci(self, process, ref_flow=None, **kwargs):
        self.check_bg()
        ref_flow = self._ensure_ref_flow(ref_flow)
        p = self._archive.retrieve_or_fetch_entity(process)
        for x in p.inventory(ref_flow=ref_flow):
            yield x

    def sys_lci(self, node, demand, **kwargs):
        raise NotImplementedError

    def bg_lcia(self, process, query_qty, ref_flow=None, **kwargs):
        p = self._archive.retrieve_or_fetch_entity(process)
        lci = self.lci(p, ref_flow=ref_flow)
        res = query_qty.do_lcia(lci, locale=p['SpatialScope'], **kwargs)
        """
        if self.privacy > 0:
            return res.aggregate('*', entity_id=p.link)
        """
        return res



from ...implementations import IndexImplementation


class IlcdIndexImplementation(IndexImplementation):
    def lcia_methods(self, **kwargs):
        self._archive.load_lcia(load_all_flows=None)
        for l in self._archive.search('quantity', **kwargs):
            if l.is_lcia_method:
                yield l



import unittest

from ..quantities import LcQuantity
from ...archives import BasicArchive, Qdb


from ...lcia_engine import IPCC_2007_GWP

mass_uuid = '93a60a56-a3c8-11da-a746-0800200b9a66'


class QuantitiesTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(QuantitiesTest, cls).setUpClass()
        cls.I = Qdb.new()
        cls.Q = BasicArchive.from_file(IPCC_2007_GWP)
        cls.gwp = cls.Q['Global Warming Air']

    def test_default_conversion(self):
        a = LcQuantity.new('Floogles', 'fl')
        self.assertTrue(a.has_property('UnitConversion'))
        self.assertEqual(a.convert('fl'), 1.0)

    def test_retrieve_quantity(self):
        self.assertIsNotNone(self.gwp)
        self.assertEqual(self.gwp.entity_type, 'quantity')

    def test_is_lcia_method(self):
        self.assertTrue(self.gwp.is_lcia_method)

    def test_convert(self):
        mass = self.I[mass_uuid]
        self.assertAlmostEqual(mass.convert('lb av'), 0.453592, 6)
        self.assertAlmostEqual(mass.convert(to='lb av'), 2.20462262, 6)

    def test_convert_missing_refunit(self):
        q = self.I.make_interface('quantity').get_canonical('electric energy')
        self.assertTrue(q.has_property('UnitConversion'))
        self.assertNotIn(q.unit, q['UnitConversion'])
        self.assertEqual(q.convert(), 1.0)
        self.assertEqual(q.convert(to='MJ'), 3.6)

    def test_valid_make_ref(self):
        gr = self.gwp.make_ref(self.Q.query)
        self.assertIsInstance(gr.name, str)


if __name__ == '__main__':
    unittest.main()



"""CATALOG_ROOT specifies the local folder that stores the reference catalog
It must either:
 - be specified in an environment variable os.environ['ANTELOPE_CATALOG_ROOT'],
 - or as a file in the local directory named antelope_catalog_root

 Used as a default location for testing and benchmarking.
"""

import os
import warnings

ENVVAR = 'ANTELOPE_CATALOG_ROOT'

try:
    CATALOG_ROOT = os.environ[ENVVAR]
    print('Using specified %s=%s' % (ENVVAR, CATALOG_ROOT))
except KeyError:
    CATALOG_ROOT = '/data/LCI/cat-demo/'  # persistent
    warnings.warn('environ %s missing; using BK default: %s' % (ENVVAR, CATALOG_ROOT))




"""
This provider needs some work:
 - the open_workbook call should not happen in the constructor, but in _load_all
 - then the provider should be marked 'static'
 - that would enable the init to be instantaneous, and the load to occur through normal channels in LcResource
 - that would also allow the loaded result to be cached in JSON
 = model: Ecoinvent LCIA

This suggests that maybe there should be a versioned LCIA base class-- probably in the data_sources hierarchy and not
the providers hierarchy though...
On that topic, it is simply foolhardy to try to build a version-agnostic provider class in the ABSENCE of multiple
versions (as is the case with TRACI)
"""

from ..xl_dict import xl_dict
from ...archives import BasicArchive
from ...entities import LcQuantity, LcFlow
from ...characterizations import DuplicateCharacterizationError

from .q_info import ns_uuid_21 as t_uuid
from .quantity import Traci21QuantityImplementation, transform_numeric_cas, CAS_regexp, q_info
from .index import Traci21IndexImplementation


def transform_string_cas(string_cas):
    return int(''.join([x for x in filter(lambda y: y != '-', string_cas)]))


class EmptyRow(Exception):
    pass


class Traci21Factors(BasicArchive):
    _ns_uuid_required = True

    def __init__(self, source, ref=None, sheet_name='Substances', mass_quantity=None, ns_uuid=t_uuid, **kwargs):
        if ref is None:
            ref = '.'.join(['local', 'traci', '2', '1', 'spreadsheet'])
        super(Traci21Factors, self).__init__(source, ref=ref, ns_uuid=ns_uuid, **kwargs)

        print('Loading workbook %s' % self.source)
        self._serialize_dict['sheet_name'] = sheet_name

        self._rows = xl_dict(self.source, sheet_name)

        self._methods = dict()  # store column-to-method mapping
        self._l_f = set()  # track which flows are loaded
        self._l_c = set()  # keep track of which methods are loaded
        for col, val in q_info.items():
            q = self._create_lcia_method(val)
            self._methods[col] = q

        if mass_quantity is None:
            self._mass = self._create_quantity('Mass', 'kg')
        else:
            self.add(mass_quantity)
            self._mass = mass_quantity

    def lcia_method_iter(self):
        for k, v in self._methods.items():
            yield v

    def make_interface(self, iface):
        if iface == 'quantity':
            return Traci21QuantityImplementation(self)
        elif iface == 'index':
            return Traci21IndexImplementation(self)
        else:
            return super(Traci21Factors, self).make_interface(iface)

    @staticmethod
    def _flow_key(flowable, compartment):
        return ', '.join([flowable, compartment])

    def _create_quantity(self, name, unitstring):
        q = self[name]
        if q is None:
            q = LcQuantity(name, Name=name, ReferenceUnit=self._create_unit(unitstring)[0])
            self.add(q)
        return q

    def _create_lcia_method(self, qinfo):
        q = self._create_quantity(qinfo.Category, qinfo.RefUnit)
        q['Method'] = 'TRACI 2.1'
        q['Category'] = qinfo.Category
        q['Indicator'] = qinfo.RefUnit
        return q

    def check_methods(self, col):
        if col in self._methods:
            return self._methods[col]
        return None

    def _create_flow(self, row):
        try:
            ext_ref = row['Substance Name'].lower()  # self._flow_key(flowable, compartment)
        except AttributeError:
            raise EmptyRow
        if ext_ref == '':
            raise EmptyRow
        f = self[ext_ref]
        if f is None:
            cas = transform_numeric_cas(row['Formatted CAS #'])
            f = LcFlow(ext_ref, Name=ext_ref, ReferenceQuantity=self._mass,
                       CasNumber=cas or '')
            self.add(f)
        return f

    def _add_flowable(self, row):
        try:
            flow = self._create_flow(row)
        except EmptyRow:
            return
        if flow.name in self._l_f:
            return

        for col, i in q_info.items():
            q = self.check_methods(col)
            try:
                cf = float(row[col])
            except (ValueError, TypeError):
                continue
            if cf == 0.0:
                continue
            self._char_from_flow_compartment_method(flow, i.Compartment, q, cf)
        self._l_f.add(flow.name)
        return flow

    def _add_column(self, col):
        if col in self._l_c:
            return
        if col in q_info:
            method = self._methods[col]
            comp = q_info[col].Compartment
            for i, row in self.iterrows():
                cf = row[col]
                try:
                    cf = float(cf)
                except (ValueError, TypeError):
                    continue
                if cf == 0.0:
                    continue
                try:
                    flow = self._create_flow(row)
                except EmptyRow:
                    continue
                self._char_from_flow_compartment_method(flow, comp, method, cf)
        self._l_c.add(col)

    def _fetch(self, entity, **kwargs):
        raise AttributeError('Cannot fetch -- this error should never occur')

    def _load_all(self):
        for i, row in self._rows.iterrows():
            self._add_flowable(row)

    def iterrows(self):
        for i, row in self._rows.iterrows():
            yield i, row

    """
    Quantity interface
    made the job difficult by electing not to make the archive static.  Trade 2 sec of load time every use for ~2
    hours of dev time. Pays off only if I load TRACI factors >3600 times.
    """

    def _char_from_flow_compartment_method(self, flow, cm, q, cf):
        cx = self.tm.add_context(cm, origin=self.ref)
        try:
            self.tm.add_characterization(flow.link, flow.reference_entity, q, cf, context=cx)
        except DuplicateCharacterizationError:
            pass

    def add_method_and_compartment(self, method=None, compartment=None):
        for col, val in q_info.items():
            q = self.check_methods(col)
            if method is None or q is method:
                cm = val.Compartment
                if compartment is None or cm == compartment:
                    self._add_column(col)

    def row_for_key(self, key):
        if CAS_regexp.match(key):
            key = transform_string_cas(key)
            for i, row in self._rows.iterrows():
                if row['CAS #'] == key:
                    yield row
        else:
            key = key.lower()
            for i, row in self._rows.iterrows():
                if row['Substance Name'].lower() == key:
                    yield row



# from .archive import Archive
# from .ecospold import EcospoldArchive
# from .ilcd import IlcdArchive


# import importlib

# from lcatools.archives import LcArchive, archive_from_json

try:
    import lxml
    from .ilcd import IlcdArchive, IlcdLcia
    from .ecospold2 import EcospoldV2Archive
    from .ecospold import EcospoldV1Archive

    _prov = ['IlcdArchive', 'IlcdLcia', 'EcospoldV2Archive', 'EcospoldV1Archive']
except ImportError:
    _prov = []

from .ecoinvent_lcia import EcoinventLcia
from .openlca_jsonld import OpenLcaJsonLdArchive
from .traci import Traci21Factors

PROVIDERS = _prov + ['EcoinventLcia', 'OpenLcaJsonLdArchive', 'Traci21Factors']

''' # this has all been folded into archive.__init__
class ArchiveError(Exception):
    pass


def create_archive(source, ds_type, catalog=None, **kwargs):
    """
    Create an archive from a source and type specification.
    :param source:
    :param ds_type:
    :param catalog: required to identify upstream archives, if specified
    :param kwargs:
    :return:
    """
    if ds_type.lower() == 'json':
        a = archive_from_json(source, factory=_provider_factory, catalog=catalog, **kwargs)
    else:
        cls = _provider_factory(ds_type)
        a = cls(source, **kwargs)
    return a


def _provider_factory(ds_type):
    """
    Returns an archive class
    :param ds_type:
    :return:
    """
    ds_type = ds_type.lower()
    init_map = {
        'lcarchive': LcArchive,
        'ilcdarchive': IlcdArchive,
        'ilcd': IlcdArchive,
        'ilcdlcia': IlcdLcia,
        'ilcd_lcia': IlcdLcia,
        'ecospoldv1archive': EcospoldV1Archive,
        'ecospold': EcospoldV1Archive,
        'ecospoldv2archive': EcospoldV2Archive,
        'ecospold2': EcospoldV2Archive,
        'ecoinventspreadsheet': EcoinventSpreadsheet,
        'ecoinventlcia': EcoinventLcia,
        'ecoinvent_lcia': EcoinventLcia,
        'openlcajsonldarchive': OpenLcaJsonLdArchive,
        'openlca': OpenLcaJsonLdArchive,
        'openlca_jsonld': OpenLcaJsonLdArchive,
        'olca': OpenLcaJsonLdArchive,
        'olca_jsonld': OpenLcaJsonLdArchive,
        'traci2': Traci21Factors,
        'traci': Traci21Factors,
        'traci21factors': Traci21Factors,
        'v1_client': AntelopeV1Client,
        'antelope_v1_client': AntelopeV1Client,
        'antelopev1client': AntelopeV1Client,
        'antelopev1': AntelopeV1Client
    }
    try:
        init_fcn = init_map[ds_type]
        return init_fcn
#        'foregroundarchive': ForegroundArchive.load,
#        'foreground': ForegroundArchive.load
    except KeyError:
        try:
            mod = importlib.import_module('.%s' % ds_type, package='antelope_%s' % ds_type)
        except ImportError as e:
            raise ArchiveError(e)
        return mod.init_fcn


def archive_from_json(fname, static=True, catalog=None, **archive_kwargs):
    """
    :param fname: JSON filename
    :param catalog: [None] necessary to retrieve upstream archives, if specified
    :param static: [True]
    :return: an ArchiveInterface
    """
    j = from_json(fname)
    archive_kwargs['quiet'] = True
    archive_kwargs['static'] = static

    if 'prefix' in j.keys():
        archive_kwargs['prefix'] = j['prefix']

    if 'nsUuid' in j.keys():
        archive_kwargs['ns_uuid'] = j['nsUuid']

    if j['dataSourceType'] == 'EcoinventSpreadsheet':
        archive_kwargs['internal'] = bool(j['internal'])
        archive_kwargs['version'] = j['version']

    if 'dataSourceReference' in j:
        # old style
        source = j['dataSourceReference']
    else:
        # new style
        source = j['dataSource']

    try:
        a = archive_factory(source, j['dataSourceType'], **archive_kwargs)
    except KeyError:
        raise ValueError('Unknown dataSourceType %s' % j['dataSourceType'])

    if 'upstreamReference' in j:
        print('**Upstream reference encountered: %s\n' % j['upstreamReference'])
        if catalog is not None:
            try:
                upstream = catalog.get_archive(j['upstreamReference'])
                a.set_upstream(upstream)
            except KeyError:
                print('Upstream reference not found in catalog!')
                a.init_args['upstreamReference'] = j['upstreamReference']
            except ValueError:
                print('Upstream reference is ambiguous!')
                a.init_args['upstreamReference'] = j['upstreamReference']
        else:
            a.init_args['upstreamReference'] = j['upstreamReference']

    a.load_from_dict(j, jsonfile=fname)
    return a
'''



import unittest
from synonym_dict.compartments.test_compartments import CompartmentContainer
from ..contexts import Context, ContextManager, InconsistentSense, ProtectedTerm
from antelope.interfaces.iindex import InvalidSense
from ..lcia_engine.lcia_engine import DEFAULT_CONTEXTS, NUM_DEFAULT_CONTEXTS


class ContextTest(CompartmentContainer.CompartmentTest):
    def test_sense(self):
        c = Context('emissions to air', sense='sink')
        self.assertEqual(c.sense, 'Sink')
        with self.assertRaises(InvalidSense):
            Context('emissions to Mars', sense='extraterrestrial')
        with self.assertRaises(InconsistentSense):
            Context('resources from urban air', parent=c, sense='source')

    def test_elementary(self):
        c = Context('eMissions', sense='sink')
        c1 = Context('emissions to boot', parent=c)
        d = Context('emulsions', sense='sink')
        self.assertTrue(c1.elementary)
        self.assertFalse(d.elementary)

    def test_parent(self):
        c = Context('emissions', sense='sink')
        d = Context('emissions to air', parent=c)
        e = Context('emissions to urban air', parent=d)
        f = Context('emissions to rural air', 'emissions from high stacks', parent=d)
        self.assertSetEqual(set(i for i in d.subcompartments), {e, f})
        self.assertListEqual([str(k) for k in c.self_and_subcompartments], ['emissions', 'emissions to air',
                                                                            'emissions to rural air',
                                                                            'emissions to urban air'])
        self.assertEqual(e.sense, 'Sink')


class ContextManagerTest(CompartmentContainer.CompartmentManagerTest):
    def _test_class(self, ignore_case=True):
        if ignore_case is False:
            self.skipTest('skipping case sensitive test')
        else:
            return ContextManager()

    def setUp(self):
        self.cm = ContextManager()

    _water_dict_objects = 3  # one higher since we start with both emissions + resources

    def test_add_from_dict(self):
        self._add_water_dict()
        self.assertEqual(str(self.cm['to water']), 'water emissions')
        self.assertEqual(self.cm['to water'].sense, 'Sink')

    def test_retrieve_by_tuple(self):
        self._add_water_dict()
        w = self.cm['to water']
        self.assertIs(w, self.cm[('emissions', 'water emissions')])

    def test_relative_add(self):
        self._add_water_dict()
        uw = self.cm['to water']
        ud = self.cm.add_compartments(['to water', 'lake water'])
        self.assertIs(uw, ud.parent)
        self.assertListEqual(ud.as_list(), ['Emissions', 'water emissions', 'lake water'])

    def test_merge_inconsistent_sense(self):
        d = [
            {
                "name": "to water",
                "parent": "Emissions"
            },
            {
                "name": "fresh water",
                "synonyms": [
                    "freshwater"
                ],
                "parent": "to water"
            },
            {
                "name": "ground-",
                "parent": "fresh water"
            },
            {
                "name": "from ground",
                "synonyms": [
                    "ground-",
                ],
                "parent": "Resources"
            }
        ]
        for k in d[:3]:
            self.cm._add_from_dict(k)
        with self.assertRaises(InconsistentSense):
            self.cm._add_from_dict(d[3])

    def _add_water_context(self):
        self._add_water_dict()
        c = self.cm.add_compartments(('to water', 'to groundwater'))
        self.cm.add_synonym(c, 'ground-')

    def test_unspecified(self):
        c = self.cm.add_compartments(['emissions', 'water', 'unspecified'])
        self.assertEqual(c.name, 'to water')
        self.assertIs(self.cm['to water, unspecified'], c)

    def test_retrieve_nonspecific(self):
        self.assertIs(self.cm.__getitem__('undefined'), self.cm._null_entry)
        self.assertIs(self.cm[None], self.cm._null_entry)

    def test_distinct_lineage(self):
        fc = self.cm.add_compartments(['fuels', 'coal', 'bituminous'])
        hc = self.cm.add_compartments(['heat', 'coal', 'pulverized'])  #  conflict='attach' is default
        self.assertIs(fc.parent, hc.parent)  # in a context manager, ambiguous specification can screw you up; 'rename' would avoid this

    def test_inconsistent_lineage(self):
        """
        context manager uses attach, then rename to avoid inconsistent lineage problems
        :return:
        """
        self._add_water_context()
        with self.assertRaises(InconsistentSense):
            self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='attach')
        fix = self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='rename')
        self.assertEqual(fix.name, 'from water, ground-')

    def test_inconsistent_lineage_match(self):
        """
        When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
        regex-based predictive guessing (reckless)
        :return:
        """
        self.skipTest('Too hard to replicate this case in ContextManager')

    def test_inconsistent_lineage_rename(self):
        self._add_water_context()
        c = self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='rename')
        self.assertEqual(c.name, 'from water, ground-')

    def test_inconsistent_lineage_skip(self):
        """
        When an intermediate descendant conflicts, we can either raise the exception (cautious) or do some clever
        regex-based predictive guessing (reckless)
        :return:
        """
        self._add_water_context()
        rw = self.cm.add_compartments(('resources', 'from water'))
        fw = self.cm.add_compartments(['resources', 'water', 'ground-'], conflict='skip')

        self.assertIs(fw, rw)

    def test_elementary(self):
        self.assertEqual(len(list(self.cm.objects)), 2)
        n = self.cm.add_compartments(['elementary flows', 'emissions', 'air', 'urban air'])
        self.assertListEqual(n.as_list(), ['elementary flows', 'Emissions', 'to air', 'urban air'])
        self.assertTrue(n.elementary)
        self.assertIn(self.cm['Emissions'], self.cm['Elementary flows'].subcompartments)
        self.assertEqual(len(list(self.cm.objects)), 5)

    def test_skip_nonspecific_spec(self):
        ew = self.cm.add_compartments(['household items', 'furniture'])
        dw = self.cm.add_compartments(['furniture', 'unspecified', 'droll'])
        self.assertIs(dw.parent, ew)
        fw = self.cm.add_compartments(['furniture', 'unspecified', 'serious'])
        self.assertIs(dw.parent, fw.parent)
        self.assertIs(self.cm['furniture, unspecified'], ew)
        self.assertIs(self.cm.__getitem__('unspecified'), self.cm._null_entry)

    def test_protected(self):
        """
        Protected terms are 'air', 'water', and 'ground'- if these are added as subcompartments to compartments with
        non-None sense, they are modified to e.g. 'to air' or 'from air' appropriate to the sense.

        We cannot actually enforce ProtectedTerm because it breaks all data sources- but we should not add protected
        synonyms to existing contexts. thus they will be directionless- unless a hint is supplied
        Really these are disallowed on *Public Query* so they should not be used but do not need to be prevented
        :return:
        """
        # with self.assertRaises(ProtectedTerm):
        #     self.cm.add_compartments(['air'])
        k = self.cm.add_compartments(('emissions', 'to air'))
        self.cm.add_synonym(k, 'secret.data:air')
        g = Context('air')
        with self.assertRaises(KeyError):  # or ProtectedTerm??
            self.cm.__getitem__(g)
        g.add_origin('secret.data')
        self.assertIs(self.cm.find_matching_context(g), k)
        self.assertIs(self.cm[g], k)
        with self.assertRaises(KeyError):
            self.cm.__getitem__('air')

    def test_as_list(self):
        r_l = ['elementary flows', 'resources', 'water', 'subterranean']
        e_l = ['elementary flows', 'emissions', 'water', 'subterranean']
        r = self.cm.add_compartments(r_l)
        e = self.cm.add_compartments(e_l)
        self.assertEqual(r.name, 'subterranean')
        self.assertEqual(e.name, 'to water, subterranean')
        self.assertListEqual(r.as_list(), ['elementary flows', 'Resources', 'from water', 'subterranean'])
        self.assertListEqual(e.as_list(), ['elementary flows', 'Emissions', 'to water', 'subterranean'])


class DefaultContextsTest(unittest.TestCase):
    def setUp(self):
        self.cm = ContextManager(source_file=DEFAULT_CONTEXTS)

    def test_load(self):
        self.assertEqual(len(self.cm), NUM_DEFAULT_CONTEXTS)
        self.assertSetEqual({k.name for k in self.cm.top_level_compartments}, {'Emissions', 'Resources'})

    def test_matching_compartment(self):
        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('resources', 'water', 'CA', 'CA-QC'))
        fx.add_origin('dummy.test')
        cx = self.cm.find_matching_context(fx)
        self.assertEqual(cx.sense, 'Source')
        self.assertIs(cx.top(), self.cm['Resources'])
        self.assertListEqual(cx.as_list(), ['Resources', 'from water'])  # superfluous information trimmed
        self.assertIs(self.cm['dummy.test:CA-QC'], cx)

    def test_context_hint(self):
        self.cm.add_context_hint('dummy.test', 'air', 'to air')
        tgt = self.cm['to air']
        self.assertIs(self.cm['dummy.test:air'], tgt)

    def test_prepend_elementary(self):
        """
        By default, emissions and resources are top-level. but if client data uses 'elementary flows' category,
        it should be installed above via 'attach'
        :return:
        """
        tw = self.cm.add_compartments(['Emissions', 'to water', 'fresh water'])
        self.assertEqual(len(tw.seq), 3)
        ew = self.cm.add_compartments(['Elementary flows', 'fresh water'])
        self.assertIs(ew, tw)
        self.assertEqual(ew.top().name, 'Elementary flows')

    '''
    def test_matching_sublineage_alt(self):
        """
        This was originally going to test the disregarded functionality- only to realize that does not get triggered
        by find_matching_context.. without that, it is more or less the same as the test below it
        :return:
        """
        self.cm.add_context_hint('dummy.test', 'air', 'to air')
        tgt = self.cm['to air']
        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('Elementary Flows', 'NETL Coal Elementary Flows', 'air', 'unspecified'))
        fx.add_origin('dummy.test')
        self.assertEqual(fx.name, 'air, unspecified')  # just throwing this in there
        self.assertIs(self.cm.find_matching_context(fx), tgt)
        # self.assertIn('Elementary Flows', self.cm.disregarded_terms)
    '''

    def test_matching_sublineage_with_hint(self):
        tgt = self.cm['from ground']

        foreign_cm = ContextManager()
        fx = foreign_cm.add_compartments(('Elementary Flows', 'NETL Coal Elementary Flows', 'NETL Elementary Flows',
                                          ' [Resources] ', 'ground'))
        # context gets its origin added when used in a Characterization or an Exchange- for now we do it manually
        fx.add_origin('dummy.test')

        self.assertIs(self.cm.find_matching_context(fx), self.cm._null_entry)

        self.cm.add_context_hint('dummy.test', '[resources]', 'Resources')
        self.assertIs(self.cm['dummy.test:[resources]'], tgt.parent)

        self.assertIs(self.cm.find_matching_context(fx), tgt)
        self.assertIs(self.cm['dummy.test:ground'], tgt)

    def test_match_tuple(self):
        res = self.cm['Resources']
        ing = self.cm['in ground']  # the true name is 'from ground'; 'in ground' is a synonym
        self.assertIs(ing.parent, res)
        self.assertIs(self.cm['Resources', 'in ground'], ing)


if __name__ == '__main__':
    unittest.main()


import unittest
from collections import namedtuple

from ..local import make_config, check_enabled


RefStats = namedtuple('RefStats', ('proc', 'elem', 'flowables'))

test_refs = {'local.ecoinvent.3.4.apos': RefStats(13290, 4078, 1184)}

_debug = False

if __name__ == '__main__':
    _run_ecoinvent = check_enabled('ecoinvent')
else:
    _run_ecoinvent = check_enabled('ecoinvent') or _debug

if _run_ecoinvent:
    from .test_aa_local import cat
    cfg = make_config('ecoinvent')


@unittest.skipIf(~_run_ecoinvent, 'Ecoinvent test skipped')
class EcoinventDataSourceTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        ar = []
        for ref in test_refs.keys():
            res = next(cfg.make_resources(ref))
            if ref not in cat.references:
                cat.add_resource(res)
            res.check(cat)
            res.archive.load_flows()
            ar.append(res.archive)
        cls.ea = tuple(ar)

    def test_nproc(self):
        for ar in self.ea:
            stats = test_refs[ar.ref]
            self.assertEqual(ar.count_by_type('process'), stats.proc, ar.ref)

    def test_nelem(self):
        for ar in self.ea:
            stats = test_refs[ar.ref]
            self.assertEqual(ar.count_by_type('flow'), stats.elem, ar.ref)

    def test_flowables(self):
        for ar in self.ea:
            stats = test_refs[ar.ref]
            self.assertEqual(len([k for k in ar.tm.flowables()]), stats.flowables, ar.ref)


if __name__ == '__main__':
    if _run_ecoinvent:
        unittest.main()



import unittest
from datetime import datetime

from ..entity_store import EntityStore
from ...entities import LcQuantity, LcUnit
from ...from_json import from_json
from .test_base import test_file


phony_source = '/path/to/phony/source.gz'
phony_ref = 'local.path.to.phony.source'
phony_ref_bumped = 'local.path.to.phony.source.bumped'


class ArchiveInterfaceTest(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls._ar = EntityStore(phony_source)

    def test_ref(self):
        self.assertEqual(self._ar.ref, phony_ref)
        self.assertEqual(self._ar.names[phony_source], phony_ref)

    def test_new_ref(self):
        now = datetime.now().strftime('%Y%m%d')
        self.assertEqual(self._ar.construct_new_ref('bumped'), '.'.join([phony_ref_bumped, now]))

    def test_nsuuid(self):
        """
        Ensure that an entity with a directly-specified UUID gets its NSUUID assigned.
        :return:
        """
        j = from_json(test_file)
        ar = EntityStore(test_file, ns_uuid=j['initArgs']['ns_uuid'])
        ar._entity_types = ('quantity', )

        q_j = next(k for k in j['quantities'] if k['referenceUnit'] == 'l')
        unit = LcUnit(q_j.pop('referenceUnit'))
        nsuuid = ar._ref_to_nsuuid(unit.unitstring)
        self.assertEqual(nsuuid, '21d34f33-b0af-3d82-9bef-3cf03e0db9dc')
        q = LcQuantity(q_j['externalId'], referenceUnit=unit, Name=q_j['Name'], entity_uuid=q_j['entityId'])
        self.assertNotEqual(q.uuid, nsuuid)
        self.assertTrue(q.validate())
        self.assertNotIn(nsuuid, ar)
        self.assertNotIn(q.uuid, ar)
        ar._add(q, q.external_ref)
        self.assertNotIn(q.uuid, ar)  # this is only done in BasicArchives
        self.assertIn(nsuuid, ar)



if __name__ == '__main__':
    unittest.main()



import json
import os
from collections import defaultdict

from antelope import UnknownOrigin
from ..lc_resource import LcResource


class ResourceNotFound(Exception):
    pass


class LcCatalogResolver(object):
    """
    The resolver maintains a collection of resources, and translates semantic references into physical archives.
    The Catalog supplies a request and a level requirement
     It also acts as a factory for those resources, so when a request is provided, it is answered with a live archive.

     Then the Catalog turns that into a static archive and keeps a list of it. The catalog also keeps a separate
     list of foreground foregrounds (which are not static; which contain fragments). These can be converted into static
     archives by turning the fragments into processes.


    """
    def __init__(self, resource_dir):
        self._resource_dir = resource_dir
        if not os.path.exists(resource_dir):
            os.makedirs(resource_dir)
        self._resources = defaultdict(list)
        self.index_resources()

    @property
    def references(self):
        """
        Generates pairs: reference, list of supported interfaces
        :return:
        """
        for k, v in self._resources.items():
            ints = set()
            for r in v:
                for t in r.interfaces:
                    ints.add(t)
            yield k, sorted(list(ints))

    @property
    def sources(self):
        seen = set()
        for k, v in self._resources.items():
            for res in v:
                if res.source not in seen:
                    seen.add(res.source)
                    yield res.source

    def _update_semantic_ref(self, ref):
        path = os.path.join(self._resource_dir, ref)
        try:
            resources = LcResource.from_json(path)
        except json.JSONDecodeError:
            print('Skipping Invalid resource file %s' % path)
            # os.remove(path)
            return
        self._resources[ref] = resources

    def index_resources(self):
        for res in os.listdir(self._resource_dir):
            self._update_semantic_ref(res)

    def add_resource(self, resource, store=True):
        """
        Add a resource to the resolver's list.  By default, save the resource permanently as a file in resources dir.
        :param resource:
        :param store: [True] if False, add the resource to memory only
        :return:
        """
        if resource.exists(self._resource_dir) or self.has_resource(resource):
            # do nothing
            print('Resource already exists')
            return
        if store:
            resource.write_to_file(self._resource_dir)
        self._resources[resource.reference].append(resource)

    def has_resource(self, resource):
        s = resource.serialize()
        return any(k.matches(s) for k in self._resources[resource.reference])

    def new_resource(self, ref, source, ds_type, store=True, **kwargs):
        new_res = LcResource(ref, source, ds_type, **kwargs)
        self.add_resource(new_res, store=store)
        return new_res

    def delete_resource(self, resource):
        ref = resource.reference
        res = self._resources[ref]
        if resource not in res:
            raise ResourceNotFound('(ref: %s)' % ref)
        res.remove(resource)
        self._write_or_delete_resource_file(ref, res)
        if len(res) == 0:
            self._resources.pop(ref)

    def known_source(self, source):
        try:
            next(self.resources_with_source(source))
        except StopIteration:
            return False
        return True

    def resources_with_source(self, source):
        for ref, ress in self._resources.items():
            for r in ress:
                if r.source == source:
                    yield r

    def is_permanent(self, resource):
        return resource.exists(self._resource_dir)

    def resolve(self, req, interfaces=None, strict=False):
        """
        Fuzzy resolver returns all references that match the request and have equal or greater specificity.
        'uslci.clean' will match queries for 'uslci' but not for 'uslci.original' or 'uslci.clean.allocated'.
        However, 'uslci.clean.allocated' will match a query for 'uslci.clean'
        :param req:
        :param interfaces: could be a single interface specification or a list
        :param strict: [False] if true, only yields interface for which req matches ref
        :return:
        """
        terms = req.split('.')
        origin_found = False
        for ref, res_list in self._resources.items():
            if strict:
                if ref != req:
                    continue
            if ref.split('.')[:len(terms)] == terms:
                origin_found = True
                for res in res_list:
                    if res.satisfies(interfaces):
                        yield res
        if not origin_found:
            raise UnknownOrigin(req)

    def get_resource(self, ref=None, iface=None, source=None, strict=True, include_internal=True):
        """
        The purpose of this function is to allow a user to retrieve a resource by providing enough information to
        identify it uniquely.  If strict is True (default), then parameters are matched exactly and more than one
        match raises an exception. If strict is False, then origins are matched approximately and the first
        (lowest-priority) match is returned.

        The convention is that no two resources should have the same source, so if a source is provided then the
        output of resources_with_source() is used.  Otherwise, the output of resolve() is used.  Internal resources
        (indexes and archives) are never returned. [WHY?]
        :return: a single LcResource
        """
        if source is not None:
            _gen = self.resources_with_source(source)
        else:
            _gen = self.resolve(ref, interfaces=iface, strict=strict)
        if include_internal:
            matches = sorted([r for r in _gen], key=lambda x: x.priority)
        else:
            matches = sorted([r for r in _gen if not r.internal], key=lambda x: x.priority)
        if len(matches) > 1:
            if strict:
                for k in matches:
                    for i in k.interfaces:
                        print('%s:%s [priority %d]' % (k.source, i, k.priority))
                raise ValueError('Ambiguous matches for supplied parameters\nref: %s iface: %s source: %s' %
                                 (ref, iface, source))
        elif len(matches) == 0:
            raise ResourceNotFound('ref:%s iface:%s source=%s' % (ref, iface, source))
        return matches[0]

    def _write_or_delete_resource_file(self, ref, resources):
        j = [k.serialize() for k in resources if k.exists(self._resource_dir)]
        if len(j) == 0:
            os.remove(os.path.join(self._resource_dir, ref))
            return
        with open(os.path.join(self._resource_dir, ref), 'w') as fp:
            json.dump({ref: j}, fp)

    def write_resource_files(self):
        for ref, resources in self._resources.items():
            self._write_or_delete_resource_file(ref, resources)
